import functools

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.middleware import csrf
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.template.backends.django import reraise
from django.template.base import Origin
from django.template.context import BaseContext
from django.utils.encoding import smart_text
from django.utils.functional import SimpleLazyObject, cached_property
from django.utils.module_loading import import_string

from ryzom import html


class Ryzom(BaseEngine):

    app_dirname = 'components'

    @staticmethod
    @functools.lru_cache()
    def get_default():
        """
        This is required for preserving historical APIs that rely on a
        globally available, implicitly configured engine such as:

        >>> from django.template import Context, Template
        >>> template = Template("Hello {{ name }}!")
        >>> context = Context({'name': "world"})
        >>> template.render(context)
        'Hello world!'
        """
        from django.template import engines

        ryzom_engines = [engine for engine in engines.all()
                         if isinstance(engine, Ryzom)]
        if len(ryzom_engines) == 1:
            return ryzom_engines[0]
        elif len(ryzom_engines) == 0:
            raise ImproperlyConfigured(
                "No Ryzom backend is configured.")
        else:
            raise ImproperlyConfigured(
                "Multiple Ryzom backends are configured. "
                "You must select one explicitly.")

    def __init__(self, params):
        params = params.copy()
        options = params.pop('OPTIONS').copy()
        self.app_dirname = options.pop("app_dirname", "components")
        super().__init__(params)

        self.debug = options.pop("debug", settings.DEBUG)
        self._context_processors = options.pop("context_processors", [])
        self.components_module = options.pop("components_module",
                                             "ryzom.components.django")

    def get_template(self, template_name):
        if template_name in html.templates:
            return Template(html.templates[template_name], self)

        try:
            return Template(template_name, self)
        except TemplateDoesNotExist as exc:
            reraise(exc, self)

    @cached_property
    def context_processors(self):
        return tuple(import_string(path) for path in self._context_processors)


class Template:

    def __init__(self, template, backend):
        if callable(template):
            # NOTE: Jinja2 chokes on non-string template name values, so
            # calls in this form require the using='ryzom' keyword argument.
            self.template = template
        else:
            try:
                self.template = import_string(template)
            except ImportError as exc:
                raise TemplateDoesNotExist(template, backend=self) from exc
            except Exception as exc:
                raise TemplateSyntaxError(template) from exc
        self.backend = backend
        self.name = f'{self.template.__module__}.{self.template.__name__}'
        self.origin = Origin(
            name=self.name,
            template_name=self.name  # TODO: No searching of app_dirs yet.
        )

    def render(self, context=None, request=None):  # noqa: C901
        from django.utils.safestring import mark_safe
        try:
            from jinja2.utils import Markup
        except ImportError:
            Markup = None

        if context is None:
            context = {}
        if request is not None:
            def _get_val():
                token = csrf.get_token(request)
                if token is None:
                    return 'NOTPROVIDED'
                else:
                    return smart_text(token)

            context["csrf_token"] = SimpleLazyObject(_get_val)

            # Support for django context processors
            for processor in self.backend.context_processors:
                context.update(processor(request))

        if self.backend.debug:
            from django.test import signals

            # Define a django-like context to imitate the multi-
            # layered context object. This is mainly for apps like
            # django-debug-toolbar that are tightly coupled to django's
            # internal implementation of context.
            if not isinstance(context, BaseContext):
                class CompatibilityContext(dict):
                    @property
                    def dicts(self):
                        return [self]

                context = CompatibilityContext(context)

            signals.template_rendered.send(sender=self,
                                           template=self,
                                           context=context)

        html = self.template().render(**context)
        if Markup:
            html = Markup(html)
        return mark_safe(html)



"""
Render Django forms using ryzom components.
"""
import logging
from collections.abc import Iterable

from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.html import format_html
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _

from ..backends.ryzom import Ryzom
from .components import (Div, Input, Label, Li, Optgroup, Option, Select, Text,
                         Textarea, Ul)

logger = logging.getLogger(__name__)


class Factory:
    """ Return the ryzom class required to render the ~django.forms.Widget.
        Initialise with a dotted module path or the template engine
        settings are used by default.
        The fallback to defaults can be disabled to force full override.

    :param str module: A dotted module path to custom components.
    :param bool fallback: Use default components, if available.
    :return: A Factory instance that will import the required Component.
    """
    def __init__(self, module=None, fallback=True):
        self.default_module = Ryzom.get_default().components_module
        self.module = module or self.default_module
        self.fallback = fallback

    def __call__(self, widget):
        cls = f'{self.module}.{type(widget).__name__}'
        try:
            cls = import_string(cls)
        except (ImportError,) as exc:  # noqa: F841
            if self.fallback and (self.module != self.default_module):
                default_cls = f'{self.default_module}.{type(widget).__name__}'
                try:
                    default_cls = import_string(default_cls)
                    logger.debug(f'Ryzom - using default widget for: {cls}')
                    cls = default_cls
                except (ImportError,) as exc:  # noqa: F841
                    pass
            # If requested widget is not found, use a basic custom widget.
            if isinstance(cls, str):
                logger.debug(f'Ryzom - widget not found: {cls}.')
                cls = f'{self.module}.TextInput'
                cls = import_string(cls)
        return cls

    @classmethod
    def as_component(self, widget):
        widget_type = type(widget).__name__
        widget_type = f'{self.default_module}.{widget_type}'
        try:
            cls = import_string(widget_type)
        except (ImportError,) as exc:  # noqa: F841
            raise NotImplementedError(
                f'Widget class {widget_type} not found.'
            )
        return cls


class Static:
    '''Return a static url for an app asset.

    :param str src: The app path of the asset.
    '''
    def __init__(self, src):
        self._data = src
        self.url = staticfiles_storage.url(src)

    def __str__(self):
        return self.url


class TextInput(Input):
    def __init__(self, widget):
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
            'type': widget['type']
        })
        if widget['value'] is not None:
            attrs['value'] = widget['value']

        super().__init__(**attrs)


class NumberInput(TextInput):
    pass


class EmailInput(TextInput):
    pass


class URLInput(TextInput):
    pass


class PasswordInput(TextInput):
    pass


class HiddenInput(TextInput):
    pass


class MultiWidget(Div):
    """ Return a list of widgets of the correct types.
        NOTE: Adds an extra div tag as a container for the widgets.
    """
    def __init__(self, multi_widget):
        content = []
        factory = Factory()
        for widget in multi_widget.subwidgets:
            attrs = widget['attrs']
            attrs.update({
                'name': widget['name'],
            })
            if widget['type'] is not None:
                attrs['type'] = widget['type']
            if widget['value'] is not None:
                attrs['value'] = widget['value']
            cls = factory(widget)
            component = cls(widget)
            content.extend(
                component if isinstance(component, Iterable) else [component]
            )
        super().__init__(*content, **attrs)


class MultipleHiddenInput(MultiWidget):
    """ Return a list of hidden widgets. """
    def __init__(self, multi_widget):
        super().__init__(multi_widget)


class FileInput(TextInput):
    pass


class ClearableFileInput():
    # TODO: Code ClearableFileInput()
    pass


class Textarea(Textarea):
    def __init__(self, widget):
        content = []
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
        })
        if widget['value'] is not None:
            content.append(
                Text(widget['value'])
            )

        super().__init__(*content, **attrs)


class DateTimeBaseInput(TextInput):
    pass


class DateInput(DateTimeBaseInput):
    pass


class DateTimeInput(DateTimeBaseInput):
    pass


class TimeInput(DateTimeBaseInput):
    pass


class CheckboxInput(TextInput):
    pass


class ChoiceWidget():
    # not directly called
    pass


class SelectOption(Option):
    def __init__(self, widget):
        attrs = widget['attrs']
        attrs.update({
            'value': widget['value'],
        })
        content = [Text(widget['label'])]
        super().__init__(*content, **attrs)


class Select(Select):
    def __init__(self, widget):
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
        })
        group_content = []
        for group_name, group_choices, group_index in widget['optgroups']:
            option_content = []
            for option in group_choices:
                option_content.append(
                    SelectOption(option)
                )
            if group_name:
                group_attrs = {
                    'label': group_name,
                }
                group_content.append(
                    Optgroup(*option_content, **group_attrs))
            else:
                group_content.extend(option_content)
        super().__init__(*group_content, **attrs)


class NullBooleanSelect(Select):
    pass


class SelectMultiple(Select):
    pass


class InputOption(Label):
    """ Return either an input element or a label wrapped around an input.
        Current style doesn't allow different element tags to be returned
        from one component so return the wrap_label version as default.
    """
    def __init__(self, widget):
        attrs = widget['attrs']
        if widget['wrap_label']:
            if attrs['id']:
                label_attrs = {
                    'for': attrs['id']
                }
            super().__init__(
                *[TextInput(widget),
                  Text(widget['label'])
                  ],
                **label_attrs
            )
        else:
            raise NotImplementedError
            # Use TextInput() if the label is not required.


class MultipleInput(Ul):
    def __init__(self, widget):
        attrs = widget['attrs']
        radio_attrs = {}
        _id = attrs['id']
        if _id:
            radio_attrs.update({
                'id': _id
            })
        if 'class' in attrs:
            radio_attrs.update({
                'class': attrs['class']
            })
        group_content = []
        for group, options, index in widget['optgroups']:
            option_content = []
            for option in options:
                option_content.append(
                    Li(*[InputOption(option) if option['wrap_label']
                         else TextInput(option)
                         ])
                )
            if group:
                group_attrs = {}
                if _id:
                    group_attrs['id'] = f'{_id}_{index}'
                group_content.append(
                    Li(
                       Text(group),
                       Ul(*option_content, **group_attrs),
                    )
                )
            else:
                group_content.extend(
                    option_content
                )

        super().__init__(*group_content, **radio_attrs)


class RadioSelect(MultipleInput):
    pass


class CheckboxSelectMultiple(MultipleInput):
    pass


class SplitDateTimeWidget(MultiWidget):
    pass


class SplitHiddenDateTimeWidget(SplitDateTimeWidget):
    pass


class SelectDateWidget(MultiWidget):
    pass


class NonFieldErrors(Ul):
    """ Render non-field errors. """
    def __init__(self, form):
        content = []
        for error in form.non_field_errors():
            content.append(
                Li(Text(error))
            )
        super().__init__(
            *content,
            **{"class": "errorlist nonfield"}
        )


class HiddenErrors(Ul):
    """ Render hidden field errors. """
    def __init__(self, form):
        content = []
        for field in form.hidden_fields():
            if field.errors:
                for error in field.errors:
                    error_text = _('(Hidden field %(name)s) %(error)s') % \
                                 {'name': field.name, 'error': str(error)}
                    content.append(
                        Li(Text(error_text))
                    )
        super().__init__(
            *content,
            **{"class": "errorlist"}
        )


class HiddenFields(Div):
    """ Render hidden fields. """
    def __init__(self, form):
        content = []
        for field in form.hidden_fields():
            content.append(
                Field(field)
            )
        super().__init__(
            *content,
            **{"class": "hidden"}
        )


class FieldErrors(Ul):
    """ Render field errors. """
    def __init__(self, errors):
        content = []
        for error in errors:
            content.append(
                Li(Text(error))
            )
        super().__init__(
            *content,
            **{"class": "errorlist"}
        )


class HelpText(Ul):
    """ Render field help text. """
    def __init__(self, help_text):
        content = []
        content.append(
            Li(Text(help_text))
        )
        super().__init__(
            *content,
            **{"class": "helptext"}
        )


class Field(Div):
    """Render a Django field using ryzom components and return an AST.

    Prepare the widget attrs, field label and context then render the
    field using ryzom components.
    Code adapted from ~django.forms.BoundField.as_widget().

    :param ~django.forms.BoundField field: The field being rendered.
    :param widget: A widget to override the default for the field.
    :type widget: ~django.forms.Widget or None
    :param attrs: Optional widget attributes.
    :type attrs: dict or None
    :param bool only_initial: A flag to render only initial dynamic values.
    :return: An AST representing the rendered field.
    :rtype: list
    """
    def __init__(self, field, widget=None, attrs=None, only_initial=False,  # noqa: C901 E501
                 label_suffix=None):
        widget = widget or field.field.widget
        if field.field.localize:
            widget.is_localized = True
        attrs = attrs or {}
        attrs = field.build_widget_attrs(attrs, widget)
        if field.auto_id and 'id' not in widget.attrs:
            attrs.setdefault('id', field.html_initial_id
                             if only_initial else field.auto_id)

        context = widget.get_context(
            name=field.html_initial_name if only_initial else field.html_name,
            value=field.value(),
            attrs=attrs,
        )
        widget_context = context['widget']
        content = []
        html_class_attrs = {}
        errors = field.form.error_class(field.errors)
        css_classes = field.css_classes()
        if css_classes:
            html_class_attrs = {"class": css_classes}

        # For MuiCheckboxInput
        label = label_tag = ''
        if field.label:
            label = field.label
            label_suffix = (label_suffix if label_suffix is not None
                            else field.field.label_suffix
                            if field.field.label_suffix is not None
                            else field.form.label_suffix)
            if label_suffix and label[-1] not in _(':?.!'):
                label_with_suffix = format_html('{}{}', label, label_suffix)
            label_tag = field.label_tag(field.label)

        widget_context['label_tag'] = label_tag
        widget_context['label'] = label_with_suffix
        widget_context['label_no_suffix'] = label

        ComponentCls = Factory()(widget)
        if label:
            # MUICSS may embed <label> after <input> in a containing div.
            if not getattr(ComponentCls, 'embed_label', False):
                content.append(
                    Text(label_tag)
                )
        component = ComponentCls(widget_context)
        content.extend(
            component if isinstance(component, Iterable) else [component]
        )

        if errors:
            content.append(FieldErrors(errors))
        if field.help_text:
            content.append(HelpText(field.help_text))
        super().__init__(
            *content,
            **html_class_attrs
        )


class VisibleFields(Div):
    """Render the regular Django fields of a form using ryzom components
    and return an AST.

    :param ~django.forms.Form: The form being rendered.
    :return: An AST representing the rendered fields.
    :rtype: list
    """
    def __init__(self, form):
        content = []
        for field in form.visible_fields():
            content.append(Field(field))
        super().__init__(
            *content
        )


class Form(Div):
    """Render a complete Django form using ryzom components and return an AST.

    :param context: Dict-like object containing a ~django.forms.Form to be
                    rendered.
    :return: An AST representing the rendered fields.
    :rtype: list
    """
    def __init__(self, context=None):
        if context is None:
            context = {}
        self.orig_context = self.context = context
        self.prepare()
        super().__init__(
            *self.content
        )

    def hidden_errors(self):
        for h in self.form.hidden_fields():
            if h.errors:
                return True

    def prepare(self, extra_context=None):
        self.content = []  # Clear any previous rendering.
        if extra_context is not None:
            # Reset context and apply extras.
            self.context = self.orig_context.update(extra_context)
        self.form = self.context.get('form', None)
        if self.form is None:
            return

        if self.form.errors:
            self.content.append(
                Ul(Li(Text(_("Please fix any errors in this form."))),
                   **{"class": "errorlist"}
                   )
            )
        # form.non_field_errors
        if self.form.non_field_errors():
            self.content.append(NonFieldErrors(self.form))
        if self.hidden_errors():  # Local method.
            self.content.append(HiddenErrors(self.form))
        # form.visible_fields
        if self.form.visible_fields():
            self.content.append(VisibleFields(self.form))
        # form.hidden_fields
        if self.form.hidden_fields():
            self.content.append(HiddenFields(self.form))
        """
        # DEBUG: helper message
        if settings.DEBUG:
            ryzom_engine = engines['ryzom']
            self.content.append(
                Text(
                    f'ryzom {ryzom_engine.components_prefix}'
                    f' Form {self.form.__class__.__name__}'
                )
            )
        """



import functools

from django.middleware.csrf import get_token
from django.utils.safestring import SafeString, mark_safe
from django.utils.html import escape

from ryzom.html import *
from .bundle import CSSBundle, JSBundle


def component_html(path, *args, **kwargs):
    try:
        from jinja2.utils import Markup
    except ImportError:
        Markup = None
    # todo: replace with importlib
    import cli2
    ComponentCls = cli2.Node(path).target
    component = ComponentCls(*args, **kwargs)
    html = component.to_html()

    if Markup:
        html = Markup(html)
    return mark_safe(html)


to_html = Component.to_html

def component_to_html(self, *args, **kwargs):
    if self.tag == 'text':
        if not isinstance(self.content, SafeString):
            return escape(f'{self.content}')
    return to_html(self, *args, **kwargs)

Component.to_html = component_to_html


class ErrorList(Ul):
    def __init__(self, *content, **attrs):
        super().__init__(
            *[Li(e) for e in content]
        )


class HiddenFields(CList):
    def __init__(self, *content, **attrs):
        super().__init__(*content, **attrs)


class Html(Html):
    scripts = [JSBundle()]
    stylesheets = [CSSBundle()]


class CSRFInput(Input):
    def __init__(self, request):
        super().__init__(
            type='hidden',
            name='csrfmiddlewaretoken',
            value=get_token(request)
        )



from django import forms

from ryzom import html

from .html import ErrorList, HiddenFields


widget_templates = dict()

def widget_template(*names):
    global widget_templates
    def decorator(component):
        for name in names:
            widget_templates[name] = component
        return component
    return decorator


def boundfield_to_component(bf, **attrs):
    try:
        template = widget_templates[bf.field.widget.template_name]
    except KeyError:
        return html.Text(str(bf))
    else:
        return template.from_boundfield(bf, **attrs)
forms.BoundField.to_component = boundfield_to_component


def boundfield_to_html(bf, **context):
    return bf.to_component().to_html(**context)
forms.BoundField.to_html = boundfield_to_html


def form_to_component(form):
    content = []
    if non_field_errors := form.non_field_errors():
        error_list = form.non_field_error_component
        content.append(error_list(*non_field_errors))

    if form.hidden_fields():
        content.append(HiddenFields(*form.hidden_fields()))
    # if form.hidden_errors():  # Local method.
    #     form.content.append(HiddenErrors(form.form))

    for bf in form.visible_fields():
        content.append(bf.to_component())

    return html.CList(str(form.media), *content)
forms.BaseForm.non_field_error_component = ErrorList
forms.BaseForm.to_component = form_to_component


def form_to_html(self, **context):
    form_component = self.to_component()
    html = form_component.to_html(**context)
    self.scripts = form_component.scripts
    self.stylesheets = form_component.stylesheets
    return html
forms.BaseForm.to_html = form_to_html


def context_attrs(widget_context, extra=None):
    # extract attrs from a widget context
    attrs = dict()
    attrs.update(dict(
        name=widget_context['name'],
        value=widget_context['value'],
    ))
    attrs.update(widget_context['attrs'])
    if 'type' in widget_context:
        attrs['type'] = widget_context['type']
    attrs.update(extra or {})
    return attrs


def boundfield_context(bf, attrs=None):
    # copy of BoundField.as_widget() with get_context instead of widget.render()
    widget = bf.field.widget
    if bf.field.localize:
        widget.is_localized = True
    attrs = attrs or {}
    attrs = bf.build_widget_attrs(attrs, widget)
    if bf.auto_id and 'id' not in widget.attrs:
        attrs.setdefault('id', bf.auto_id)
    return widget.get_context(
        name=bf.html_name,
        value=bf.value(),
        attrs=attrs,
    )


def widget_context(bf, attrs=None):
    return boundfield_context(bf, attrs)['widget']


def widget_attrs(bf, attrs=None):
    return context_attrs(widget_context(bf), attrs)


def field_kwargs(bf):
    return dict(
        name=bf.name,
        label=bf.label,
        value=bf.value(),
        help_text=bf.help_text,
        errors=bf.errors,
    )



'''
Defines the ryzom View class and the main index view
'''
from asgiref.sync import async_to_sync
from django import http

from py2js.renderer import JS, autoexec
from ryzom.components import Component
from ryzom_django_channels.models import (Client, Publication, Registration,
                                          Subscription)


class ReactiveMixin:
    def get_token(view):
        user = view.request.user
        try:
            client = Client.objects.create(user=user)
        except ValueError:
            client = Client.objects.create()

        view.client = client

        def js_set_token():
            window.token = token
            ws_connect()

        return autoexec(JS(js_set_token, dict(token=client.token)))


class RegisterManager:
    def __init__(self, queryset):
        self.queryset = queryset

    def update(self, content):
        from channels.layers import get_channel_layer
        channel = get_channel_layer()
        for registration in self.queryset:
            content.id = registration.subscriber_id
            content.parent = registration.subscriber_parent
            channel_name = registration.client.channel
            if channel_name:
                async_to_sync(channel.send)(channel_name, {
                    'type': 'handle.ddp',
                    'params': {
                        'type': 'changed',
                        'instance': content.to_obj()
                    }
                })


def register(register_name):
    queryset = Registration.objects.filter(name=register_name)
    return RegisterManager(queryset)



from django.utils.safestring import mark_safe
from ryzom.html import *


class MDCLink(A):
    pass


class MDCIcon(Icon):
    attrs = {
        'class': 'material-icons',
        'aria-hiddem': 'true'
    }


class MDCTextButton(Button):
    def __init__(self, text, icon=None, **kwargs):
        content = [Span(cls='mdc-button__ripple')]
        if icon:
            content.append(MDCIcon(icon))
        content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            cls='mdc-button',
            **kwargs,
        )


class MDCButton(Button):
    attrs = {'class': 'mdc-button'}

    def __init__(self, text=None, icon=None, **kwargs):
        if icon and isinstance(icon, str):
            content = [MDCIcon(icon)]
        elif icon:
            content = [icon]
        else:
            content = []

        content.append(Span(cls='mdc-button__ripple'))
        if text:
            content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            **kwargs
        )


class MDCButtonRaised(MDCButton):
    attrs = {'addcls': 'mdc-button--raised'}


class MDCButtonOutlined(MDCButton):
    attrs = {'addcls': 'mdc-button--outlined'}


class MDCButtonLabelOutlined(Label):
    def __init__(self, text, p=True, icon=None):
        black = 'black-button' if p else ''
        content = [Span(cls='mdc-button__ripple')]
        if icon:
            content.append(MDCIcon(icon))
        content.append(Span(text, cls='mdc-button__label'))
        super().__init__(
            *content,
            cls=f'mdc-button mdc-button--outlined {black}'
        )


class MDCTextInput(Input):
    def __init__(self, name, **attrs):
        attrs['name'] = name
        attrs.setdefault('type', 'text')
        attrs.setdefault('class', 'mdc-text-field__input')
        super().__init__(**attrs)


class MDCLabel(Span):
    attrs = {'class': 'mdc-floating-label'}


class MDCTextRipple(Span):
    def __init__(self):
        super().__init__(**{'class': 'mdc-text-field__ripple'})


class MDCLineRipple(Span):
    def __init__(self):
        super().__init__(**{'class': 'mdc-line-ripple'})


class MDCTextFieldFilled(Label):
    attrs = {
        'class': 'mdc-text-field mdc-text-field--filled',
    }

    def __init__(self, label='', input_id='', label_id='', **kwargs):
        content = [
            MDCTextRipple(),
            MDCLabel(label, label_id),
            MDCTextInput(input_id, **kwargs),
            MDCLineRipple(),
        ]
        super().__init__(*content, **self.attrs)


class MDCNotchOutline(Span):
    attrs = {'class': 'mdc-notched-outline'}

    def __init__(self, *content, **attrs):
        super().__init__(
            Span(cls='mdc-notched-outline__leading'),
            Span(
                *content,
                cls='mdc-notched-outline__notch',
            ),
            Span(cls='mdc-notched-outline__trailing'),
            **attrs,
        )


class MDCField(Div):
    VERTICAL_MARGIN = '32px'
    style = f'margin-top: {VERTICAL_MARGIN}; margin_bottom: {VERTICAL_MARGIN}'

    def __init__(self, *content, name, label=None, help_text=None, value=None,
                 errors=None, **attrs):
        helper_id = f'id_{name}_helper'
        errors_id = f'id_{name}_errors'

        if errors:
            self.errors = MDCErrorList(*errors, id=errors_id)
        else:
            self.errors = ''

        if help_text:
            self.help_text = MDCHelpText(help_text, id=helper_id)
        else:
            self.help_text = ''

        super().__init__(*content, self.errors, self.help_text, **attrs)


class MDCTextFieldOutlined(MDCField):
    def __init__(self, html_input, label=None, help_text=None, errors=None, **attrs):
        self.html_input = html_input
        self.html_input.attrs.addcls = 'mdc-text-field__input'

        name = self.html_input.attrs.name
        input_id = f'id_{name}'
        label_id = f'id_{name}_label'
        helper_id = f'id_{name}_helper'
        errors_id = f'id_{name}_errors'

        floating_label = Span(label, id=label_id, cls='mdc-floating-label')
        notch_outline = MDCNotchOutline(floating_label)
        self.label = Label(
            notch_outline,
            self.html_input,
            id=label_id,
            cls='mdc-text-field mdc-text-field--outlined',
            data_mdc_auto_init='MDCTextField',
        )
        self.html_input.attrs.aria_labelledby = label_id

        value = self.html_input.attrs.get('value', '')
        if value not in ('', None):
            # float label because there is a value
            self.label.attrs.addcls = 'mdc-text-field--label-floating'
            floating_label.attrs.addcls = 'mdc-floating-label--float-above'
            notch_outline.attrs.addcls = 'mdc-notched-outline--notched'

        if errors:
            self.label.attrs.addcls = 'mdc-text-field--invalid'
            self.html_input.attrs.aria_describedby = errors_id
            self.html_input.attrs.aria_controls = errors_id
        elif help_text:
            self.html_input.attrs.aria_describedby = helper_id
            self.html_input.attrs.aria_controls = helper_id

        super().__init__(self.label, name=name,
                         help_text=help_text,
                         errors=errors, **attrs)


class MDCTextareaFieldOutlined(MDCTextFieldOutlined):
    def __init__(self, textarea, label=None, help_text=None, errors=None):
        super().__init__(
            textarea,
            label=label,
            help_text=help_text,
            errors=errors,
        )
        # decorate textarea with a resizer
        self.label.content[1] = Span(textarea, cls='mdc-text-field__resizer')
        self.label.attrs.addcls = 'mdc-text-field--textarea'


class MDCFormField(Div):
    attrs = {'class': 'mdc-form-field'}

    def __init__(self, *content, **kwargs):
        super().__init__(*content, **self.attrs, **kwargs)


class MDCFileField(Div):
    def __init__(self, html_input, label=None, help_text=None, errors=None, **attrs):
        self.btn = MDCButtonLabelOutlined(label, False)
        self.input_id = html_input.attrs['id']
        self.btn.attrs['for'] = self.input_id
        self.selected_text = Span('No file selected')
        self.label_id = self.selected_text.id
        super().__init__(
            Span(
                html_input,
                style='display:block;width:0;height:0;overflow:hidden'
            ),
            self.selected_text,
            self.btn
        )

    def set_update_name(input_id, label_id):
        def update_name(event):
            file_name = event.target.value
            label = getElementByUuid(label_id)
            label.innerHTML = file_name or 'No file selected'

        document.querySelector('#'+input_id).addEventListener('change', update_name)

    def py2js(self):
        self.set_update_name(self.input_id, self.label_id)


class MDCSplitDateTime(Div):
    def __init__(self, label='', input_id='', label_id='', **kwargs):
        name = kwargs.pop('name')
        kwargs.pop('type', None)
        value = kwargs.pop('value', None)
        date = value.date() if value else ''
        time = value.time() if value else ''

        super().__init__(
            MDCTextFieldOutlined(
                label='Date',
                input_id='date_' + name + '_input',
                name=name + '_0',
                value=date,
                type='date',
                **kwargs),
            MDCTextFieldOutlined(
                label='Time',
                input_id='time_' + name + '_input',
                name=name + '_1',
                value=time,
                type='time',
                **kwargs),
        )

    def set_error(self, error):
        helper = MDCTextFieldHelperLine(error, 'alert')
        label = self.content[1]
        label.attrs['class'] += ' mdc-text-field--invalid'
        label.attrs['aria-describedby'] = helper.id
        label.attrs['aria-controls'] = helper.id

        self.content.append(helper)


class MDCTextFieldHelperText(Div):
    attrs = {
        'class': ' '.join([
            'mdc-text-field-helper-text',
            'mdc-text-field-helper-text--persitent',
        ]),
    }

class MDCTextFieldHelperLine(Div):
    attrs = {
        'aria-hidden': 'true',
        'cls': 'mdc-text-field-helper-line',
    }


class MDCVerticalMargin(Div):
    style = 'margin-top: 18px; margin-bottom: 18px'


class MDCList(Div):
    attrs = {'class': 'mdc-list', 'data-mdc-auto-init': 'MDCList'}


class MDCListItem(Li):
    attrs = {'class': 'mdc-list-item'}

    def __init__(self, *content, icon=None, ripple=True, **attrs):
        if icon and not isinstance(icon, Component):
            icon = MDCIcon(icon, addcls='mdc-list-item__graphic')

        super().__init__(
            Span(cls='mdc-list-item__ripple') if ripple else None,
            icon,
            Span(*content, cls='mdc-list-item__text'),
            **attrs,
        )


class MDCSnackBar(Div):
    def __init__(self, msg, status='success', delay=0):
        self.delay = delay
        super().__init__(
            Div(
                Div(
                    msg,
                    cls='mdc-snackbar__label',
                    **{'aria-atomic': 'false'}
                ),
                Div(
                    Button(
                        Div(cls='mdc-button__ripple'),
                        Span('OK', cls='mdc-button__label'),
                        type='button',
                        cls='mdc-button mdc-snackbar__action'
                    ),
                    cls='mdc-snackbar__actions',
                    **{'aria-atomic': 'true'}
                ),
                cls='mdc-snackbar__surface',
                role='status',
                **{'aria-relevant': 'addition'}
            ),
            cls='mdc-snackbar',
            **{'data-mdc-auto-init': 'MDCSnackbar'}
        )

    def open_snack(elem):
        sb = new.mdc.snackbar.MDCSnackbar(elem)
        sb.open()

    def py2js(self):
        setTimeout(
            self.open_snack,
            self.delay,
            getElementByUuid(self.id)
        )


class MDCErrorListItem(Li):
    style = dict(list_style_type='none')


class MDCErrorList(Ul):
    style = dict(
        padding_left='10px',
        margin_top=0,
        margin_bottom=0,
        font_size='var(--mdc-typography-caption-font-size, 0.75rem)',
        color='var(--mdc-theme-error, #b00020)',
    )
    def __init__(self, *content, **attrs):
        super().__init__(*[
            MDCErrorListItem(c) if isinstance(c, str) else c
            for c in content
        ], **attrs)


class MDCHelpText(Div):
    style = dict(
        font_size='var(--mdc-typography-caption-font-size, 0.75rem)',
        color='rgba(0, 0, 0, 0.6)',
        padding_left='10px',
    )


class MDCInputCheckboxNativeControl(Input):
    attrs = {'class': 'mdc-checkbox__native-control'}


class MDCCheckboxInput(Div):
    """
    The actual input HTML element (widget).

    TODO: stop hijacking attrs like this because it enforces downstream
    boilerplate
    """
    def __init__(self, input=None, **attrs):
        attrs.setdefault('type', 'checkbox')
        super().__init__(
            cls='mdc-checkbox',
            input=input or MDCInputCheckboxNativeControl(**attrs),
            checkbox=Div(
                Component(
                    Component(
                        tag='path', fill='none',
                        d="M1.73,12.91 8.1,19.28 22.79,4.59",
                        cls='mdc-checkbox__checkmark-path'),
                    tag='svg', viewBox='0 0 24 24',
                    cls='mdc-checkbox__checkmark'
                ),
                Div(cls='mdc-checkbox__mixedmark'),
                cls='mdc-checkbox__background',
            ),
        )


class MDCCheckboxField(MDCField):
    def __init__(self, *content, name, label=None, help_text=None, value=None,
                 errors=None):
        super().__init__(
            MDCFormField(
                *content,
                Label(label or name.capitalize()),
            ),
            name=name,
            errors=errors,
            help_text=help_text,
        )

        # compensate for widget margin
        if self.errors:
            self.errors.attrs.style.margin_top = '-10px'
        elif self.help_text:
            self.help_text.attrs.style.margin_top = '-10px'


class MDCCheckboxSelectField(MDCField):
    def __init__(self, *content, name, label=None, help_text=None, value=None,
                 errors=None):
        super().__init__(
            *content,
            name=name,
            errors=errors,
            help_text=help_text,
        )

        # compensate for widget margin
        if self.errors:
            self.errors.attrs.style.margin_top = '-10px'
        elif self.help_text:
            self.help_text.attrs.style.margin_top = '-10px'


class MDCCheckboxListItem(Li):
    def __init__(self, title, id, checked=False, **kwargs):
        self.input_id = id
        if checked:
            kwargs['checked'] = True
        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            Span(
                Div(
                    Input(
                        id=id,
                        type='checkbox',
                        cls='mdc-checkbox__native-control',
                        **kwargs),
                    Div(
                        Component(
                            Component(
                                tag='path', fill='none',
                                d="M1.73,12.91 8.1,19.28 22.79,4.59",
                                cls='mdc-checkbox__checkmark-path'),
                            tag='svg', viewBox='0 0 24 24',
                            cls='mdc-checkbox__checkmark'),
                        Div(cls='mdc-checkbox__mixedmark'),
                        cls='mdc-checkbox__background'),
                    cls='mdc-checkbox'),
                cls='mdc-list-item__graphic'),
            Label(title, cls='mdc-list-item__text', **{'for': id}),
            cls='mdc-list-item',
            role='checkbox',
            **{
                'aria-checked': 'true' if checked else 'false',
                'data-list-item-of': id
            }
        )

    def click_input(event):
        event.stopPropagation()
        elem = event.target.querySelector('input')
        if elem:
            elem.click()

    def py2js(self):
        elem = getElementByUuid(self.id)
        elem.onclick = self.click_input


class MDCMultipleChoicesCheckbox(Ul):
    def __init__(self, name, choices, n=1, **kwargs):
        self.max = n
        alabel = kwargs.pop('aria-label', 'Label')
        super().__init__(
            *(MDCCheckboxListItem(
                title,
                f'{name}_input_{i}',
                name=name,
                value=value,
                **kwargs
            ) for i, title, value in choices),
            cls='mdc-list',
            role='group',
            **{'aria-label': alabel}
        )

    def update_inputs(event):
        input_list = event.currentTarget
        checked = input_list.querySelectorAll('input:checked')
        unchecked = input_list.querySelectorAll('input:not(:checked)')

        def disable(elem, pos, arr):
            elem.disabled = True
            list_item = document.querySelector(
                '[data-list-item-of="' + elem.id + '"]'
            ).classList.add('mdc-list-item--disabled')

        def enable(elem, pas, arr):
            elem.disabled = undefined
            list_item = document.querySelector(
                '[data-list-item-of="' + elem.id + '"]'
            ).classList.remove('mdc-list-item--disabled')

        if checked.length >= self.max:
            unchecked.forEach(disable)
        else:
            unchecked.forEach(enable)

    def py2js(self):
        input_list = getElementByUuid(self.id)
        input_list.max = self.max
        input_list.addEventListener('change', self.update_inputs)


class MDCSelect(Div):
    attrs = {'class': 'mdc-select mdc-select--filled', 'data-mdc-auto-init': 'MDCSelect'}

    def __init__(self, select, *content, **attrs):
        super().__init__(
            mark_safe('''
            <div class="mdc-select__anchor" role="button" aria-haspopup="listbox"
                  aria-labelledby="demo-pagination-select" tabindex="0">
              <span class="mdc-select__selected-text-container">
                <span id="demo-pagination-select" class="mdc-select__selected-text">10</span>
              </span>
              <span class="mdc-select__dropdown-icon">
                <svg
                    class="mdc-select__dropdown-icon-graphic"
                    viewBox="7 10 10 5">
                  <polygon
                      class="mdc-select__dropdown-icon-inactive"
                      stroke="none"
                      fill-rule="evenodd"
                      points="7 10 12 15 17 10">
                  </polygon>
                  <polygon
                      class="mdc-select__dropdown-icon-active"
                      stroke="none"
                      fill-rule="evenodd"
                      points="7 15 12 10 17 15">
                  </polygon>
                </svg>
              </span>
              <span class="mdc-notched-outline mdc-notched-outline--notched">
                <span class="mdc-notched-outline__leading"></span>
                <span class="mdc-notched-outline__trailing"></span>
              </span>
            </div>

            <div class="mdc-select__menu mdc-menu mdc-menu-surface mdc-menu-surface--fullwidth" role="listbox">
              <ul class="mdc-list">
            '''),
            *[
                mark_safe(f'''
                <li class="mdc-list-item {'mdc-list-item--selected" aria-selected="true' if option.attrs.get('selected', False) else ''}" role="option" data-value="{option.attrs.value}">
                  <span class="mdc-list-item__text">{option.attrs.value}</span>
                </li>
                ''')
                for option in select.content
            ],
            mark_safe('''
              </ul>
            </div>
            '''),
            **attrs,
        )


class MDCOption(Li):
    def __init__(self, index, **choice):
        extra_attrs = dict()

        selected = choice.get('selected', False)
        if selected:
            extra_attrs['addcls'] = 'mdc-list-item--selected'
            extra_attrs['aria-selected'] = 'true'

        if index == '0':
            extra_attrs['tabindex'] = 0

        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            Span(
                choice['label'] if choice['value'] else '',
                cls='mdc-list-item__text'
            ),
            data_value=choice['value'],
            cls='mdc-list-item',
            role='option',
            **extra_attrs,
        )


class MDCNamedOptgroup(Div):
    def __init__(self, name, choices):
        super().__init__(
            Ul(
                H6(name, cls='mdc-list-group__subheader'),
                *(MDCOption(**choice) for choice in choices),
                cls='mdc-list'
            ),
            cls='mdc-list-group'
        )


class MDCOptgroup(CList):
    def __init__(self, name, choices, index):
        if name:
            super().__init__(MDCNamedOptgroup(name, choices))
        else:
            super().__init__(
                *(MDCOption(**choice) for choice in choices)
            )


class MDCSelectAnchor(Div):
    def __init__(self, label, selected=None, **attrs):
        required = attrs.pop('required', None)

        super().__init__(
            Span(
                Span(cls='mdc-notched-outline__leading'),
                Span(
                    label,
                    cls='mdc-notched-outline__notch'),
                Span(cls='mdc-notched-outline__trailing'),
                cls='mdc-notched-outline'),
            Span(
                selected or Span(cls='mdc-select__selected-text'),
                cls='mdc-select__selected-text-container'),
            Span(
                Svg(
                    Polygon(
                        stroke='none',
                        fill_rule='evenodd',
                        points='7 10 12 15 17 10',
                        cls='mdc-select__dropdown-icon-inactive'),
                    Polygon(
                        stroke='none',
                        fill_rule='evenodd',
                        points='7 15 12 10 17 15',
                        cls='mdc-select__dropdown-icon-active'),
                    viewBox="7 10 10 5",
                    focusable="false",
                    cls='mdc-select__dropdown-icon-graphic'),
                cls='mdc-select__dropdown-icon'),
            cls='mdc-select__anchor',
            role="button",
            aria_required='true' if required else 'false',
            aria_labelled_by=label.id,
            aria_haspopup="listbox",
            aria_expanded="false",
        )


class MDCSelectMenu(Div):
    def __init__(self, optgroups, **attrs):
        super().__init__(
            Ul(
                *(
                    MDCOptgroup(group_name, group_choices, group_index)
                    for group_name, group_choices, group_index
                    in optgroups
                ),
                cls='mdc-list'
            ),
            cls='mdc-select__menu mdc-menu mdc-menu-surface mdc-menu-surface--fullwidth',
        )


class MDCSelectOutlined(Div):
    tag = 'mdc-select-outlined'

    def __init__(self, **attrs):
        attrs.pop('template_name')
        label = attrs.pop('label', attrs.get('name', None))

        cls = 'mdc-select mdc-select--outlined'
        if required := attrs.pop('required', None):
            cls += ' mdc-select--required'

        super().__init__(
            Input(
                type='hidden',
                name=attrs['name'],
                value=attrs['value'][0] if attrs['value'] else '',
            ),
            MDCSelectAnchor(
                Span(
                    label,
                    cls='mdc-floating-label'
                ),
                **dict(required=required),
            ),
            MDCSelectMenu(**attrs),
            cls=cls,
            data_mdc_auto_init='MDCSelect',
        )

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('change', this.change.bind(this))

        def change(self, event):
            hidden = this.querySelector('input[type=hidden]')
            option = this.querySelector('[aria-selected=true]')
            hidden.value = option.dataset.value


class MDCAccordionToggle(MDCListItem):
    tag = 'mdc-accordion-toggle'

    def __init__(self, *, label=None, **context):
        super().__init__(label, icon='add', **context)

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('click', this.click.bind(this))

        def click(self, event):
            section = this.parentElement
            if section.classList.contains('active'):
                section.toggle()
            else:
                section.parentElement.closeAll()
                section.open()


class MDCAccordionMenu(Div):
    tag = 'mdc-accordion-menu'

    style = dict(
        display='block',
        overflow='hidden',
        max_height='0px',
    )

    def __init__(self, *content, **context):
        super().__init__(
            *content,
            cls='MDCAccordionMenu',
            **context
        )

    class HTMLElement:
        def connectedCallback(self):
            window.addEventListener('load', this.ready.bind(this))

        def ready(self):
            max_height = this.style.maxHeight
            this.style.transition = ''
            this.close()
            this.from_px = '0px'
            if max_height and max_height != '0px':
                this.from_px = max_height
                this.open()

        def start_layout(self):
            this.style.transition = ''
            this.from_px = this.style.maxHeight
            this.style.maxHeight = 'initial'
            this.rect = this.getBoundingClientRect()

            closest = this.parentElement.closest('mdc-accordion-menu')
            if closest:
                closest.start_layout()

        def end_layout(self):
            this.style.maxHeight = self.from_px
            this.getBoundingClientRect()
            this.style.transition='max-height 0.4s ease-out'
            this.style.maxHeight = this.rect.height

            closest = this.parentElement.closest('mdc-accordion-menu')
            if closest:
                closest.end_layout()

        def open(self):
            this.start_layout()
            this.end_layout()

        def close(self):
            this.style.maxHeight = 0


class MDCAccordionSection(MDCList):
    tag = 'mdc-accordion-section'

    def __init__(self, *content, **context):
        super().__init__(
            MDCAccordionToggle(**context),
            MDCAccordionMenu(*content, **context),
            addcls='mdc-accordion',
        )

    class HTMLElement:
        def open(self):
            this.querySelector('mdc-accordion-toggle').classList.add(
                'mdc-list-item--selected'
            )
            this.classList.add('active')
            this.querySelector('mdc-accordion-menu').open()
            this.querySelector('i').innerText = 'remove'

        def close(self):
            this.querySelector('mdc-accordion-toggle').classList.remove(
                'mdc-list-item--selected'
            )
            this.classList.remove('active')
            this.querySelector('mdc-accordion-menu').close()
            this.querySelector('i').innerText = 'add'

        def toggle(self):
            if this.classList.contains('active'):
                this.close()
            else:
                this.open()


class MDCAccordion(Div):
    tag = 'mdc-accordion'

    def __init__(self, *content, **context):
        def get_content():
            for c in content[:-1]:
                yield c
                yield Hr(cls='mdc-list-divider')
            if len(content):
                yield content[-1]

        super().__init__(
            *(c for c in get_content()),
            addcls='mdc-accordion'
        )

    class HTMLElement:
        def closeAll(self):
            sections = this.querySelectorAll('mdc-accordion-section')
            for section in sections:
                section.close()


class MDCSelectPerPage(MDCSelect):
    tag = 'mdc-select-per-page'

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('MDCSelect:change', this.change.bind(this))

        async def change(self, event):
            url = new.URL(document.location)
            if url.search.indexOf('per_page=') > 0:
                search = url.search.replace(
                    new.RegExp('per_page=[^&]*'),
                    'per_page=' + event.detail.value,
                )
            else:
                search = '?per_page=' + event.detail.value

            if document.up and document.up.visit:
                up.visit(url.pathname + search, {target: '.mdc-data-table'})
            else:
                document.location.href = url.pathname + search


class MDCIconButton(A):
    attrs = dict(cls='material-icons mdc-top-app-bar__navigation-icon mdc-icon-button')


class MdcTopAppBar(Component):
    """
    <header class="mdc-top-app-bar">
      <div class="mdc-top-app-bar__row">
        <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-start">
          <button class="material-icons mdc-top-app-bar__navigation-icon mdc-icon-button" aria-label="Open navigation menu">menu</button>
          <span class="mdc-top-app-bar__title">Page title</span>
        </section>
        <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-end" role="toolbar">
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Favorite">favorite</button>
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Search">search</button>
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Options">more_vert</button>
        </section>
      </div>
    </header>
    """
    tag = 'header'

    def __init__(self, *content, **context):
        action_items = context.pop('action_items', None)
        super().__init__(
            Div(
                # Navigation menu icon and page title.
                Component(
                    Button(
                        'menu',
                        aria_label="Open navigation menu",
                        cls='material-icons mdc-top-app-bar__navigation-icon mdc-icon-button',
                    ),
                    Span(
                        *content,
                        cls='mdc-top-app-bar__title',
                    ),
                    cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-start',
                    tag='section',
                ),
                # Actions items and menu.
                MdcAppBarActions(action_items=action_items),
                cls='mdc-top-app-bar__row',
            ),
            cls='mdc-top-app-bar app-bar',
            id='app-bar',
            data_mdc_auto_init='MDCTopAppBar',
            **context,
        )


class MdcDrawer(Component):
    """
    <aside class="mdc-drawer mdc-drawer--dismissible">
      <div class="mdc-drawer__header">
        <h3 class="mdc-drawer__title">Drawer/h3>
        <h6 class="mdc-drawer__subtitle">Demonstration</h6>
      </div>
      <div class="mdc-drawer__content">
        <nav class="mdc-list">
          <a class="mdc-list-item mdc-list-item--activated" href="#" aria-current="page">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">inbox</i>
            <span class="mdc-list-item__text">Inbox</span>
          </a>
          <a class="mdc-list-item" href="#">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">send</i>
            <span class="mdc-list-item__text">Outgoing</span>
          </a>
          <hr class="mdc-list-divider">
          <h6 class="mdc-list-group__subheader">Labels</h6>
          <a class="mdc-list-item" href="#">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">drafts</i>
            <span class="mdc-list-item__text">Drafts</span>
          </a>
        </nav>
      </div>
    </aside>
    """
    tag = 'aside'

    def __init__(self, *content, **context):
        super().__init__(
            # Header.
            Div(
                H3(
                    *context['mdc-header-title'],
                    cls="mdc-drawer__title",
                ),
                H6(
                    *context['mdc-header-subtitle'],
                    cls="mdc-drawer__subtitle",
                ),
                cls="mdc-drawer__header",
            ) if 'mdc-header-title' in context else '',
            # List.
            Div(
                *content,
                cls='mdc-drawer__content',
            ),
            cls='mdc-drawer mdc-drawer--dismissible mdc-top-app-bar--fixed-adjust',
            data_mdc_auto_init='MDCDrawer',
            **context,
        )

class MdcAppContent(Div):
    """
    <div class="mdc-drawer-app-content mdc-top-app-bar--fixed-adjust">
      <main class="main-content" id="main-content">
        App Content
      </main>
    </div>
    """
    def __init__(self, *content, **context):
        super().__init__(
            Component(
                *content,
                tag='main',
                cls='main-content',
                id='main-content',
            ),
            cls='mdc-drawer-app-content mdc-top-app-bar--fixed-adjust',
        )


class MdcDrawerHeader(Div):
    """
    <div class="mdc-drawer__header">
      <h3 id="drawer-title" class="mdc-drawer__title">Title</h3>
      <h6 id="drawer-subtitle" class="mdc-drawer__subtitle">subtitle</h6>
    </div>
    """
    def __init__(self, *content, **context):
        title = context.pop('drawer_title', '')
        subtitle = context.pop('drawer_subtitle', '')
        super().__init__(
            *content,
            H3(
                title,
                cls='mdc-drawer__title',
                id='drawer-title',
            ),
            H6(
                subtitle,
                cls='mdc-drawer__subtitle',
                id='drawer-subtitle',
            ),
            cls='mdc-drawer__header',
        )


class MdcListDivider(Text):
    """
    <hr class="mdc-list-divider">
    <h6 class="mdc-list-group__subheader">Section</h6>
    """
    def __init__(self, *content, **context):
        super().__init__(
            Hr(cls="mdc-list-divider"),
            H6(
                *content,
                cls="mdc-list-group__subheader",
                **context,
            ),
        )


class MdcIcon(Component):
    tag = 'i'

    def __init__(self, *content, **context):
        super().__init__(
            *content,
            cls='material-icons mdc-list-item__graphic',
            aria_hidden='true',
            **context,
        )


class MdcList(Component):
    """ Allow tag to be overridden via context.

    """
    def __init__(self, *content, **context):
        context.setdefault('tag', 'ul')
        context.setdefault('cls', '')
        context['cls'] = (' '.join((context['cls'], 'mdc-list'))
                          if context['cls'] else 'mdc-list')
        super().__init__(*content, **context)


class MdcNavList(MdcList):
    def __init__(self, *content, **context):
        context['tag'] = 'nav'
        super().__init__(*content, **context)


class MdcListItem(Component):
    """ MdcListItem(Text("Home"), icon='home', href='#', active=True)
    <a class="mdc-list-item mdc-list-item--activated" href="#" aria-current="page">
      <span class="mdc-list-item__ripple"></span>
      <i class="material-icons mdc-list-item__graphic" aria-hidden="true">home</i>
      <span class="mdc-list-item__text">Home</span>
    </a>
    """
    def __init__(self, *content, icon=None, **context):
        context.setdefault('tag', 'a')
        context.setdefault('aria-current', 'page')
        context.setdefault('cls', '')
        context['cls'] = (' '.join((context['cls'], 'mdc-list-item'))
                          if context['cls'] else 'mdc-list-item')
        if context.pop('active', None):
            context['cls'] += ' mdc-list-item--activated'
            context['tabindex'] = 0
        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            MdcIcon(icon) if icon else '',
            Span(*content, cls='mdc-list-item__text'),
            **context,
        )


class MdcAppBarActions(Component):
    def __init__(self, *content, **context):
        action_items = context.pop('action_items', None)
        if not action_items:
            super().__init__()
        else:
            context['tag'] = 'section'
            super().__init__(
                *[Button(
                    item[0],  # icon
                    aria_label=item[1],  # label
                    href=item[2],
                    cls='material-icons mdc-top-app-bar__action-item mdc-icon-button',
                ) for item in action_items],
                cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-end',
                role='toolbar',
                **context,
            )


class MdcButton(Div):
    def __init__(self, label=None, name=None, *content, **context):
        label = label or (name and name.capitalize())
        context.setdefault('tag', 'button')
        context.setdefault('cls', '')
        context['cls'] += 'mdc-button mdc-button--touch mdc-button--raised'
        super().__init__(
            # TODO: What tag here? Context?
            Component(
                Span(cls='mdc-button__ripple'),
                *content,
                Span(label, cls='mdc-button__label'),
                Span(cls='mdc-button__touch'),
                **context,
            ),
            cls='mdc-touch-target-wrapper',
            data_mdc_auto_init='MDCRipple',
        )


class MdcTextField(Label):
    """
    <label class="mdc-text-field mdc-text-field--filled mdc-text-field--label-floating">
      <span class="mdc-text-field__ripple"></span>
      <input class="mdc-text-field__input" type="text" aria-labelledby="my-label-id" value="Initial value">
      <span class="mdc-floating-label mdc-floating-label--float-above" id="my-label-id">
        Label in correct place
      </span>
      <span class="mdc-line-ripple"></span>
    </label>
    """
    def __init__(self, name=None, label=None, value=None, **context):
        label = label or (name and name.capitalize()) or ''
        super().__init__(
            Span(cls='mdc-text-field__ripple'),
            Input(
                name=name,
                value=value or '',
                type='text',
                aria_labelledby=f'id_{name}-label',
                cls='mdc-text-field__input',
                **context,
            ),
            Span(
                Text(label),
                id=f'{name}-label',
                cls='mdc-floating-label',
            ),
            Span(cls='mdc-line-ripple'),
            cls='mdc-text-field mdc-text-field--filled',
            data_mdc_auto_init='MDCTextField',
        )


class MDCLayoutGrid(Div):
    attrs = {'class': 'mdc-layout-grid'}


class MDCLayoutGridInner(Div):
    attrs = {'class': 'mdc-layout-grid__inner'}


class MDCLayoutGridCell(Div):
    attrs = {'class': 'mdc-layout-grid__cell'}


class MDCCard(Div):
    attrs = {'class': 'mdc-card'}


class MDCDataTable(Div):
    attrs = {'class': 'mdc-data-table', 'data-mdc-auto-init': 'MDCDataTable'}

    def __init__(self, *content, container=None, table=None,
            tbody=None, thead=None, **attrs):

        super().__init__(
            container=container or MDCDataTableContainer(
                table=table or MDCDataTableTable(
                    thead=thead,
                    tbody=tbody,
                )
            ),
            **attrs,
        )
        self.table = self.container.table
        self.tbody = self.table.tbody
        self.thead = self.table.thead


class MDCDataTableResponsive(MDCDataTable):
    sass = '''
    @media (max-width: 700px)
        .MDCDataTableResponsive
            thead
                display: none
            tr
                display: block
            td
                display: block
                text-align: right
                height: auto
            td:before
                content: attr(data-label)
                float: left
                font-weight: bold
    '''


class MDCDataTableTable(Table):
    attrs = {'class': 'mdc-data-table__table'}

    def __init__(self, *content, thead=None, tbody=None, **attrs):
        super().__init__(
            *content,
            thead=thead or MDCDataTableThead(),
            tbody=tbody or MDCDataTableTbody(),
            **attrs,
        )

    def to_html(self, *content, **context):
        if 'view' in context and getattr(context['view'], 'title', None):
            # set a default aria label if possible
            self.attrs.arial_label = context['view'].title
        return super().to_html(*content, **context)



class MDCDataTableTbody(Tbody):
    attrs = {'class': 'mdc-data-table__content'}


class MDCDataTableThead(Thead):
    def __init__(self, *content, tr=None, **attrs):
        super().__init__(*content, tr=tr or MDCDataTableTr(), **attrs)


class MDCDataTableTh(Th):
    attrs = {
        'class': 'mdc-data-table__header-cell',
        'role': 'columnheader',
        'scope': 'col',
    }


class MDCDataTableHeaderTr(Tr):
    attrs = {'class': 'mdc-data-table__header-row'}


class MDCDataTableTr(Tr):
    attrs = {'class': 'mdc-data-table__row'}


class MDCDataTableTd(Td):
    attrs = {'class': 'mdc-data-table__cell'}


class MDCDataTableContainer(Div):
    attrs = {'class': 'mdc-data-table__table-container'}

    def __init__(self, *content, table=None, **attrs):
        return super().__init__(
            table=table or MDCDataTableTable(),
            **attrs,
        )


class MDCDrawerToggle(Component):
    tag = 'mdc-drawer-toggle'

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('click', this.toggle.bind(this))

        def toggle(self):
            drawer = document.getElementById(this.attributes['data-drawer-id'].value)
            drawer = mdc.drawer.MDCDrawer.attachTo(drawer)
            drawer.open = not drawer.open


class ToggleNextElement(Component):
    tag = 'toggle-element'
    style = {
        ' svg': {
            'transition': 'all 0.5s ease',
        },
        '.open svg': {
            'transform': 'rotate(180deg)',
        }
    }

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('click', this.click.bind(this))

        def click(self, event):
            element = this.nextElementSibling
            if element.style.display == 'none':
                element.style.display = 'block'
                this.classList.add('open')
            else:
                element.style.display = 'none'
                this.classList.remove('open')


class MDCFilterField(Component):
    style = {
        ' h3': {
            'display': 'flex',
            'justify-content': 'space-between',
        }
    }

    def __init__(self, label=None, widget=None, **attrs):
        super().__init__(
            label=ToggleNextElement(
                H3(
                    Span(label),
                    Component(
                        mark_safe('<g><path d="M16.59 8.59L12 13.17 7.41 8.59 6 10l6 6 6-6z"></path> <path d="M0 0h24v24H0z" fill="none"></path></g><svg>'),
                        tag='svg',
                        width='26',
                        height='26',
                        viewBox='0 0 26 26',
                        role='presentation',
                        xmlns='http://www.w3.org/2000/svg',
                        cls='gc-icon gc-icon--expand',
                    ),
                ),
            ),
            widget=Div(
                widget,
                style={'transition': 'all 1s ease'},
            ),
            **attrs,
        )

    @classmethod
    def from_boundfield(cls, bf):
        return cls(
            label=bf.label,
            field=bf.to_component(),
        )


class MDCDataTablePagination(Div):
    attrs = {'class': 'mdc-data-table__pagination'}
    sass = '''
    @media (max-width: 700px)
        .MDCDataTablePagination [slot]
            display: block
    '''


class MDCDialogContainer(Div):
    attrs = {'class': 'mdc-dialog__container'}


class MDCDialogTitle(H2):
    attrs = {'class': 'mdc-dialog__title'}


class MDCDialogActions(Div):
    attrs = {'class': 'mdc-dialog__actions'}


class MDCDialogContent(Div):
    attrs = {'class': 'mdc-dialog__content'}


class MDCDialogSurface(Div):
    def __init__(self, title, content, actions=None, **attrs):
        extra_attrs = {}

        _title = MDCDialogTitle(title)
        _content = MDCDialogContent(content)
        _actions = actions

        if _actions is None:
            _actions = MDCDialogActions(
                    MDCDialogCloseButtonOutlined('cancel'),
                    MDCDialogAcceptButton('confirm'),
                    style='display: flex; justify-content: space-around'
                )

        super().__init__(
            _title,
            _content,
            _actions,
            cls='mdc-dialog__surface',
            role='alertdialog',
            aria_modal='true',
            aria_labelledby=_title.id,
            aria_describedby=_content.id
        )


class MDCDialogCloseButton(MDCButton):
    tag = 'a'
    attrs = {'data-mdc-dialog-action': 'close'}


class MDCDialogCloseButtonOutlined(MDCButtonOutlined):
    tag = 'a'
    attrs = {'data-mdc-dialog-action': 'close'}


class MDCDialogAcceptButton(MDCButtonRaised):
    tag = 'a'
    attrs = {'data-mdc-dialog-action': 'accept'}


class MDCDialogAcceptButtonOutlined(MDCButtonOutlined):
    tag = 'a'
    attrs = {'data-mdc-dialog-action': 'accept'}


class MDCDialogScrim(Div):
    attrs = {'class': 'mdc-dialog__scrim'}


class MDCDialog(Div):
    tag = 'mdc-dialog'
    attrs = {
        'class': 'mdc-dialog',
        'data-mdc-auto-init': 'MDCDialog'
    }

    def __init__(self, *content, **attrs):
        actions = attrs.pop('actions', None)
        super().__init__(
            MDCDialogContainer(
                MDCDialogSurface(*content, actions=actions)
            ),
            MDCDialogScrim(),
            **attrs
        )

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('MDCDialog:closing', self.handle_closing.bind(this))
            this.addEventListener('MDCDialog:closed', self.handle_closed.bind(this))
            this.addEventListener('MDCDialog:opening', self.handle_opening.bind(this))
            this.addEventListener('MDCDialog:opened', self.handle_opened.bind(this))

        def onclosing(self, event):
            pass

        def onclosed(self, event):
            pass

        def onopening(self, event):
            pass

        def onopened(self, event):
            pass

        def open(self):
            this.MDCDialog.open()

        def close(self):
            this.MDCDialog.close()

        def layout(self):
            this.MDCDialog.layout()

        def handle_closing(self, event):
            this.onclosing(event)

        def handle_closed(self, event):
            this.onclosed(event)

        def handle_opening(self, event):
            this.onopening(event)

        def handle_opened(self, event):
            this.onopened(event)


class MDCChip(Div):
    attrs = {'class': 'mdc-chip', 'role': 'row'}

    def __init__(self, *content, icon=None, **attrs):
        super().__init__(
            MDCChipRipple(),
            Gridcell(*content),
            icon=Gridcell(icon) if icon else '',
            **attrs
        )


class MDCChipRipple(Div):
    attrs = {'class': 'mdc-chip__ripple'}


class Gridcell(Span):
    attrs = {'role': 'gridcell'}


class InlineForm(Form):
    style = {
        'display': 'inline-block',
        ' .MDCField': {
            'display': 'inline-block',
        },
    }


class Body(Body):
    attrs = {'class': 'mdc-typography'}

    def py2js(self):
        mdc.autoInit()


class Html(Html):
    scripts = [
        'https://unpkg.com/material-components-web@10.0.0/dist/material-components-web.min.js',
        'https://unpkg.com/@webcomponents/webcomponentsjs@2.0.0/webcomponents-bundle.js',
        'https://cdn.polyfill.io/v2/polyfill.min.js',
        '/static/py2js.js',
    ]
    stylesheets = [
        'https://fonts.googleapis.com/icon?family=Material+Icons',
        'https://fonts.googleapis.com/css?family=Roboto:300,400,500',
        'https://unpkg.com/material-components-web@10.0.0/dist/material-components-web.min.css',
    ]
    body_class = Body



'''
Routes channels websocket incoming message to ryzom.consumers.Consumer
'''
from django.conf.urls import url
from django.core.asgi import get_asgi_application

from . import consumers

websocket_urlpatterns = [
    url(r'^ws/ddp/$', consumers.Consumer.as_asgi()),
    url(r'', get_asgi_application())
]



"""
Test helpers for Ryzom TDD.

Example:

.. code-block:: python

    def test_columns():
        table = Table(
            [dict(name='foo', test='y')],
            columns=['name'],
        )
        result = table.to_html()
        assert pretty(result) == '''
    %table
      %thead
        %tr
          %th
            name
      %tbody
        %tr
          %td
            foo
    '''.strip()


In pytest, this will display a beautiful diff in case of failure !
"""

import os
import re
import pytest  # noqa: F401
from lxml.html import HtmlElement, fromstring


def tree(html):
    return html if isinstance(html, HtmlElement) else fromstring(html)


def pretty(html, indent=None):
    indent = indent or 0
    spaces = indent * ' '
    el = tree(html)

    out = [f'{spaces}%{el.tag}\n']
    for key in sorted(el.attrib.keys()):
        if key == 'ryzom-id':
            continue
        out += [f'{spaces}  {key}={el.attrib[key]}\n']
    if el.text:
        out += [f'{spaces}  {el.text}\n']
    for child in el:
        out += pretty(child, indent + 2)

    if el.tail:
        out += [f'{spaces}{el.tail}\n']

    if not indent:
        return ''.join(out).strip()

    return out


def test_pretty():
    assert pretty('''
    <foo bar="test"><a x="Y">text1<br />text2</a></foo>
    ''') == '''%foo
  bar=test
  %a
    x=Y
    text1
    %br
    text2
'''.strip()


ryzom_id_re = re.compile(r'(?<=ryzom-id=")([^"]*)')
re_uuid = re.compile(r'"[a-z0-9]{32}"')
csrf_re = r'[<][^<]*name="csrfmiddlewaretoken"[^>]*[>]'


def assert_equals(expected, result):
    from django.test.html import parse_html
    expected = parse_html(expected)
    result = parse_html(result)
    assert result == expected


def assert_equals_fixture(name, result):
    path = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        '..',
        '..',
        'tests',
        'fixtures',
        f'{name}.html',
    ))
    result = re.sub(ryzom_id_re, '', str(result))
    result = re.sub(re_uuid, '""', str(result))
    result = re.sub(csrf_re, 'csrfmiddlewaretoken', result)
    if not os.path.exists(path):
        result = re.sub(ryzom_id_re, '', str(result))
        with open(path, 'w') as f:
            f.write(result)
        pytest.fail('Fixture created')
    with open(path, 'r') as f:
        expected = f.read()
    assert_equals(expected, result)



import importlib
import textwrap

try:
    import sass
except:
    sass = None


def to_css(selector, style):
    out = []

    maincss = []
    for style_key, style_value in style.items():
        if isinstance(style_value, dict):
            continue
        maincss.append(f'  {style_key}: {style_value};')

    if maincss:
        out = [selector + ' {', *maincss, '}']

    for sub_selector, sub_style in style.items():
        if not isinstance(sub_style, dict):
            continue
        out += to_css(f'{selector}{sub_selector}', sub_style)

    return out


def bundle(*modules):
    out = []
    done = []
    for module in modules:
        mod = importlib.import_module(module)
        for key, value in mod.__dict__.items():
            if not isinstance(value, type):
                continue
            if not hasattr(value, 'attrs'):
                continue
            if value in done:
                continue
            if sass_src := getattr(value, 'sass', None):
                if not sass:
                    raise Exception(f'Sass not installed, cannot bundle {key}')
                css = sass.compile(
                    string=textwrap.dedent(sass_src),
                    indented=True
                )
                if css not in out:
                    out.append(css)
            if 'style' in value.attrs:
                out += to_css('.' + value.__name__, value.attrs.style)
            done.append(value)
    return '\n'.join(out)



'''
Ryzom MUI CSS components.
'''
import re
from collections.abc import Iterable

from django.utils.html import conditional_escape
from django.utils.translation import gettext as _

from .components import Select as RyzSelect
from .components import *  # noqa
from .django import Factory, InputOption, SelectOption
from .django import TextInput as DjangoTextInput

"""
class Component(Component):
    def __init__(self, *args, **kwargs):
        self.context = kwargs.pop('context', {})
        super().__init__(*args, **kwargs)


class Appbar(Component):
    '''
    Appbar component

    Represents a MUI CSS <appbar>.

    :parameters: see :class:`Component`
    '''
    def __init__(self, content=None, attr=None, events=None,
                 parent='body', _id=None, context=None):
        return Div(content,
                   attr={'class': "mui-appbar"},
                   events, parent, _id)


class Button(Component):
    '''
    Button component

    Represents a MUI CSS <button>.

    :parameters: see :class:`Component`
    '''
    def __init__(self, content=None, attr=None,
                 events=None, parent='body', _id=None, context=None):
        attr = attr or {}
        cls = attr.setdefault('class', '')
        attr['class'] = f"mui-btn {cls}"
        return Button(content, attr, events, parent, _id)


class Container(Component):
    '''
    Container component

    Represents a MUI CSS <container>.

    :parameters: see :class:`Component`
    '''
    def __init__(self, content=None, attr=None, events=None,
                 parent='body', _id=None, context=None):
        return Div(content,
                   attr={'class': "mui-container"},
                   events, parent, _id)
"""


class MuiForm(Component):
    '''
    Form component

    Represents a MUI CSS <form>.

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        attrs = {'class': "mui-form"}
        super().__init__(*content, **attrs, tag='form')


class Legend(Component):
    '''
    Legend component

    Represents a MUI CSS <legend>.

    :parameters: see :class:`Component`
    '''
    def __init__(self, *content, **attrs):
        super().__init__(*content, **attrs, tag='legend')


class TextInput(Div):
    # Prevent default pre-element labelling
    embed_label = True

    def __init__(self, widget):

        div_content = []
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
            'type': widget['type'],
        })
        if widget['value'] is not None:
            attrs['value'] = widget['value']
        div_content.append(
            Input(**attrs),
        )
        if 'label_tag' in widget:
            # Radio input options won't have a label_tag.
            div_content.append(
                widget['label_tag'],
            )
        div_attrs = {'class': "mui-textfield mui-textfield--float-label"}

        super().__init__(*div_content, **div_attrs)


class NumberInput(TextInput):
    pass


class EmailInput(TextInput):
    pass


class URLInput(TextInput):
    pass


class PasswordInput(TextInput):
    pass


class HiddenInput(TextInput):
    pass


class MultiWidget(Div):
    """ Return a list of widgets of the correct types.
        NOTE: Adds an extra div tag as a container for the widgets.
    """
    def __init__(self, multi_widget):
        content = []
        factory = Factory()
        for widget in multi_widget.subwidgets:
            attrs = widget['attrs']
            attrs.update({
                'name': widget['name'],
            })
            if widget['type'] is not None:
                attrs['type'] = widget['type']
            if widget['value'] is not None:
                attrs['value'] = widget['value']
            cls = factory(widget)
            component = cls(widget)
            content.extend(
                component if isinstance(component, Iterable) else [component]
            )
        super().__init__(*content, **attrs)


class MultipleHiddenInput(MultiWidget):
    """ Return a list of hidden widgets. """
    def __init__(self, multi_widget):
        super().__init__(multi_widget)


class FileInput(TextInput):
    pass


# TODO: Code ClearableFileInput()
# class ClearableFileInput():
#     pass


class Textarea(Textarea):
    # Prevent default pre-element labelling
    embed_label = True

    def __init__(self, widget):
        content = []
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
        })
        if widget['value'] is not None:
            content.append(
                Text(widget['value']),
                widget['label_tag'],
            )
        div_content = []
        div_content.append(
            Textarea(*content, **attrs)
        )
        div_attrs = {'class': "mui-textfield mui-textfield--float-label"}

        super().__init__(*div_content, **div_attrs)


class DateTimeBaseInput(TextInput):
    pass


class DateInput(DateTimeBaseInput):
    pass


class DateTimeInput(DateTimeBaseInput):
    pass


class TimeInput(DateTimeBaseInput):
    pass


class CheckboxInput(Div):
    # Prevent default pre-element labelling
    embed_label = True

    def __init__(self, widget):
        div_content = []
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
            'type': widget['type'],
        })
        if widget['value'] is not None:
            attrs['value'] = widget['value']

        # widget['label_tag'] isn't useful here.
        div_content.append(
            Label(
                Input(**attrs),
                Text(widget.get(
                    'label', re.findall('>([^<]*)<', widget['label_tag'])[0])),
                **{'for': widget['attrs']['id']}
            )
        )
        div_attrs = {'class': "mui-checkbox"}

        super().__init__(*div_content, **div_attrs)


class ChoiceWidget():
    # not directly called
    pass


class Select(Div):
    def __init__(self, widget):
        div_content = []
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
        })
        group_content = []
        for group_name, group_choices, group_index in widget['optgroups']:
            option_content = []
            for option in group_choices:
                option_content.append(
                    SelectOption(option)
                )
            if group_name:
                group_attrs = {
                    'label': group_name,
                }
                group_content.append(
                    Optgroup(*option_content, **group_attrs))
            else:
                group_content.extend(option_content)

        div_content.append(
            RyzSelect(*group_content, **attrs),
        )
        div_attrs = {"class": "mui-select"}
        super().__init__(*div_content, **div_attrs)


class NullBooleanSelect(Select):
    pass


class SelectMultiple(Div):
    def __init__(self, widget):
        options = []
        for group_name, group_choices, group_index in widget['optgroups']:
            for option in group_choices:
                options.append(
                    SelectOption(option)
                )

        super().__init__(
            Component(
                Component(
                    *options, multiple='true', name=widget['name'],
                    tag='select', slot='select'
                ),
                Component(slot='deck', tag='div'),
                Component(
                    Component(slot='input', tag='input', type='text'),
                    slot='input',
                    tag='autocomplete-light',
                ),
                multiple='true',
                tag='autocomplete-select'
            )
        )


class MultipleInput(Ul):
    def __init__(self, widget):
        attrs = widget['attrs']
        radio_attrs = {}
        _id = attrs['id']
        if _id:
            radio_attrs.update({
                'id': _id
            })
        if 'class' in attrs:
            radio_attrs.update({
                'class': attrs['class']
            })
        group_content = []
        for group, options, index in widget['optgroups']:
            option_content = []
            for option in options:
                option_content.append(
                    Li(InputOption(option)
                           if option['wrap_label']
                           else DjangoTextInput(option)
                           )
                )
            if group:
                group_attrs = {}
                if _id:
                    group_attrs['id'] = f'{_id}_{index}'
                group_content.append(
                    Li(
                        Text(group),
                        Ul(*option_content, **group_attrs),
                        )
                )
            else:
                group_content.extend(
                    option_content
                )

        super().__init__(*group_content, **radio_attrs)


class RadioSelect(MultipleInput):
    pass


class CheckboxSelectMultiple(MultipleInput):
    pass


class SplitDateTimeWidget(MultiWidget):
    pass


class SplitHiddenDateTimeWidget(SplitDateTimeWidget):
    pass


class SelectDateWidget(MultiWidget):
    pass


class Button(Button):
    """
    {{ render_button(view.title_submit, button_type="submit",
        button_class="btn-primary") }}
    """
    def __init__(self, widget):
        """
        def render_button(
            content,
            button_type=None,
            icon=None,
            button_class="btn-default",
            size="",
            href="",
            name=None,
            value=None,
            title=None,
            extra_classes="",
            id="",
        ):

        """
        div_content = []
        attrs = widget['attrs']
        attrs.update({
            'name': widget['name'],
            'type': widget['type'],
        })
        if widget['value'] is not None:
            attrs['value'] = widget['value']
        div_content.append(
            Input(**attrs),
            widget['label'],
        )
        div_attrs = {'class': "mui-textfield mui-textfield--float-label"}

        super().__init__(*div_content, **div_attrs)


class NonFieldErrors(Ul):
    """ Render non-field errors. """
    def __init__(self, form):
        content = []
        for error in form.non_field_errors():
            content.append(
                Li(Text(error))
            )
        super().__init__(
            *content,
            **{"class": "errorlist nonfield"}
        )


class HiddenErrors(Ul):
    """ Render hidden field errors. """
    def __init__(self, form):
        content = []
        for field in form.hidden_fields():
            if field.errors:
                for error in field.errors:
                    error_text = _('(Hidden field %(name)s) %(error)s') % \
                                 {'name': field.name, 'error': str(error)}
                    content.append(
                        Li(Text(error_text))
                    )
        super().__init__(
            *content,
            **{"class": "errorlist"}
        )


class HiddenFields(Div):
    """ Render hidden fields. """
    def __init__(self, form):
        content = []
        for field in form.hidden_fields():
            content.append(
                Field(field)
            )
        super().__init__(
            *content,
            **{"class": "hidden"}
        )


class FieldErrors(Ul):
    """ Render field errors. """
    def __init__(self, errors):
        content = []
        for error in errors:
            content.append(
                Li(Text(error))
            )
        super().__init__(
            *content,
            **{"class": "errorlist"}
        )


class HelpText(Ul):
    """ Render field help text. """
    def __init__(self, help_text):
        content = []
        content.append(
            Li(Text(help_text))
        )
        super().__init__(
            *content,
            **{"class": "helptext"}
        )


class Field(Div):
    """Render a MUI field using ryzom components and return an AST.

    Prepare the widget attrs, field label and context then render the
    field using ryzom components.
    Code adapted from ~django.forms.BoundField.as_widget().

    :param ~django.forms.BoundField field: The field being rendered.
    :param widget: A widget to override the default for the field.
    :type widget: ~django.forms.Widget or None
    :param attrs: Optional widget attributes.
    :type attrs: dict or None
    :param bool only_initial: A flag to render only initial dynamic values.
    :return: An AST representing the rendered field.
    :rtype: list
    """
    def __init__(self, field, widget=None, attrs=None, only_initial=False):  # noqa: C901 E501
        widget = widget or field.field.widget
        if field.field.localize:
            widget.is_localized = True
        attrs = attrs or {}
        attrs = field.build_widget_attrs(attrs, widget)
        if field.auto_id and 'id' not in widget.attrs:
            attrs.setdefault('id', field.html_initial_id
                             if only_initial else field.auto_id)

        context = widget.get_context(
            name=field.html_initial_name if only_initial else field.html_name,
            value=field.value(),
            attrs=attrs,
        )
        widget_context = context['widget']
        content = []
        html_class_attrs = {}
        errors = field.form.error_class(field.errors)
        css_classes = field.css_classes()
        if css_classes:
            html_class_attrs = {"class": css_classes}

        # For CheckboxInput
        label_chkbox = ''
        if field.label:
            label = conditional_escape(field.label)
            label_chkbox = label
            label = field.label_tag(label) or ''
        else:
            label = ''

        widget_context['label_tag'] = label
        widget_context['label'] = label_chkbox

        ComponentCls = Factory()(widget)
        if label:
            # MUICSS may embed <label> after <input> in a containing div.
            if not getattr(ComponentCls, 'embed_label', False):
                content.append(
                    Text(label)
                )
        component = ComponentCls(widget_context)
        content.extend(
            component if isinstance(component, Iterable) else [component]
        )

        if errors:
            content.append(FieldErrors(errors))
        if field.help_text:
            content.append(HelpText(field.help_text))
        super().__init__(
            *content,
            **html_class_attrs
        )


class VisibleFields(Div):
    """Render the regular Django fields of a form using ryzom components
    and return an AST.

    :param ~django.forms.Form: The form being rendered.
    :return: An AST representing the rendered fields.
    :rtype: list
    """
    def __init__(self, form):
        content = []
        for field in form.visible_fields():
            content.append(Field(field))
        super().__init__(
            *content
        )


class Form(Div):
    """Render a complete Django form using ryzom components and return an AST.

    :param ~django.forms.Form: The form being rendered.
    :return: An AST representing the rendered fields.
    :rtype: list
    """
    def __init__(self, form):
        content = []
        # form.non_field_errors
        content.append(NonFieldErrors(form))
        content.append(HiddenErrors(form))
        # form.visible_fields
        content.append(VisibleFields(form))
        # form.hidden_fields
        content.append(HiddenFields(form))

        super().__init__(
            *content,
        )



'''
Ryzom mdc (Material Design Components) components.
'''
from .components import *  # noqa


class MdcTopAppBar(Component):
    """
    <header class="mdc-top-app-bar">
      <div class="mdc-top-app-bar__row">
        <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-start">
          <button class="material-icons mdc-top-app-bar__navigation-icon mdc-icon-button" aria-label="Open navigation menu">menu</button>
          <span class="mdc-top-app-bar__title">Page title</span>
        </section>
        <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-end" role="toolbar">
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Favorite">favorite</button>
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Search">search</button>
          <button class="material-icons mdc-top-app-bar__action-item mdc-icon-button" aria-label="Options">more_vert</button>
        </section>
      </div>
    </header>
    """
    tag = 'header'

    def __init__(self, *content, **context):
        action_items = context.pop('action_items', None)
        super().__init__(
            Div(
                # Navigation menu icon and page title.
                Component(
                    Button(
                        'menu',
                        aria_label="Open navigation menu",
                        cls='material-icons mdc-top-app-bar__navigation-icon mdc-icon-button',
                    ),
                    Span(
                        *content,
                        cls='mdc-top-app-bar__title',
                    ),
                    cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-start',
                    tag='section',
                ),
                # Actions items and menu.
                MdcAppBarActions(action_items=action_items),
                cls='mdc-top-app-bar__row',
            ),
            cls='mdc-top-app-bar app-bar',
            id='app-bar',
            data_mdc_auto_init='MDCTopAppBar',
            **context,
        )


class MdcDrawer(Component):
    """
    <aside class="mdc-drawer mdc-drawer--dismissible">
      <div class="mdc-drawer__header">
        <h3 class="mdc-drawer__title">Drawer/h3>
        <h6 class="mdc-drawer__subtitle">Demonstration</h6>
      </div>
      <div class="mdc-drawer__content">
        <nav class="mdc-list">
          <a class="mdc-list-item mdc-list-item--activated" href="#" aria-current="page">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">inbox</i>
            <span class="mdc-list-item__text">Inbox</span>
          </a>
          <a class="mdc-list-item" href="#">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">send</i>
            <span class="mdc-list-item__text">Outgoing</span>
          </a>
          <hr class="mdc-list-divider">
          <h6 class="mdc-list-group__subheader">Labels</h6>
          <a class="mdc-list-item" href="#">
            <span class="mdc-list-item__ripple"></span>
            <i class="material-icons mdc-list-item__graphic" aria-hidden="true">drafts</i>
            <span class="mdc-list-item__text">Drafts</span>
          </a>
        </nav>
      </div>
    </aside>
    """
    tag = 'aside'

    def __init__(self, *content, **context):
        super().__init__(
            # Header.
            Div(
                H3(
                    *context['mdc-header-title'],
                    cls="mdc-drawer__title",
                ),
                H6(
                    *context['mdc-header-subtitle'],
                    cls="mdc-drawer__subtitle",
                ),
                cls="mdc-drawer__header",
            ) if 'mdc-header-title' in context else '',
            # List.
            Div(
                *content,
                cls='mdc-drawer__content',
            ),
            cls='mdc-drawer mdc-drawer--dismissible mdc-top-app-bar--fixed-adjust',
            data_mdc_auto_init='MDCDrawer',
            **context,
        )

class MdcAppContent(Div):
    """
    <div class="mdc-drawer-app-content mdc-top-app-bar--fixed-adjust">
      <main class="main-content" id="main-content">
        App Content
      </main>
    </div>
    """
    def __init__(self, *content, **context):
        super().__init__(
            Component(
                *content,
                tag='main',
                cls='main-content',
                id='main-content',
            ),
            cls='mdc-drawer-app-content mdc-top-app-bar--fixed-adjust',
        )


class MdcDrawerHeader(Div):
    """
    <div class="mdc-drawer__header">
      <h3 id="drawer-title" class="mdc-drawer__title">Title</h3>
      <h6 id="drawer-subtitle" class="mdc-drawer__subtitle">subtitle</h6>
    </div>
    """
    def __init__(self, *content, **context):
        title = context.pop('drawer_title', '')
        subtitle = context.pop('drawer_subtitle', '')
        super().__init__(
            *content,
            H3(
                title,
                cls='mdc-drawer__title',
                id='drawer-title',
            ),
            H6(
                subtitle,
                cls='mdc-drawer__subtitle',
                id='drawer-subtitle',
            ),
            cls='mdc-drawer__header',
        )


class MdcListDivider(Text):
    """
    <hr class="mdc-list-divider">
    <h6 class="mdc-list-group__subheader">Section</h6>
    """
    def __init__(self, *content, **context):
        super().__init__(
            Hr(cls="mdc-list-divider"),
            H6(
                *content,
                cls="mdc-list-group__subheader",
                **context,
            ),
        )


class MdcIcon(Component):
    tag = 'i'

    def __init__(self, *content, **context):
        super().__init__(
            *content,
            cls='material-icons mdc-list-item__graphic',
            aria_hidden='true',
            **context,
        )


class MdcList(Component):
    """ Allow tag to be overridden via context.

    """
    def __init__(self, *content, **context):
        context.setdefault('tag', 'ul')
        context.setdefault('cls', '')
        context['cls'] = (' '.join((context['cls'], 'mdc-list'))
                          if context['cls'] else 'mdc-list')
        super().__init__(*content, **context)


class MdcNavList(MdcList):
    def __init__(self, *content, **context):
        context['tag'] = 'nav'
        super().__init__(*content, **context)


class MdcListItem(Component):
    """ MdcListItem(Text("Home"), icon='home', href='#', active=True)
    <a class="mdc-list-item mdc-list-item--activated" href="#" aria-current="page">
      <span class="mdc-list-item__ripple"></span>
      <i class="material-icons mdc-list-item__graphic" aria-hidden="true">home</i>
      <span class="mdc-list-item__text">Home</span>
    </a>
    """
    def __init__(self, *content, icon=None, **context):
        context.setdefault('tag', 'a')
        context.setdefault('aria-current', 'page')
        context.setdefault('cls', '')
        context['cls'] = (' '.join((context['cls'], 'mdc-list-item'))
                          if context['cls'] else 'mdc-list-item')
        if context.pop('active', None):
            context['cls'] += ' mdc-list-item--activated'
            context['tabindex'] = 0
        super().__init__(
            Span(cls='mdc-list-item__ripple'),
            MdcIcon(icon) if icon else '',
            Span(*content, cls='mdc-list-item__text'),
            **context,
        )


class MdcAppBarActions(Component):
    def __init__(self, *content, **context):
        action_items = context.pop('action_items', None)
        if not action_items:
            return ''
        context['tag'] = 'section'
        super().__init__(
            *[Button(
                item[0],  # icon
                aria_label=item[1],  # label
                href=item[2],
                cls='material-icons mdc-top-app-bar__action-item mdc-icon-button',
            ) for item in action_items],
            cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-end',
            role='toolbar',
            **context,
        )


class MdcButton(Div):
    def __init__(self, label=None, name=None, *content, **context):
        label = label or (name and name.capitalize())
        context.setdefault('tag', 'button')
        context.setdefault('cls', '')
        context['cls'] += 'mdc-button mdc-button--touch mdc-button--raised'
        super().__init__(
            # TODO: What tag here? Context?
            Component(
                Span(cls='mdc-button__ripple'),
                *content,
                Span(label, cls='mdc-button__label'),
                Span(cls='mdc-button__touch'),
                **context,
            ),
            cls='mdc-touch-target-wrapper',
            data_mdc_auto_init='MDCRipple',
        )


class MdcTextField(Label):
    """
    <label class="mdc-text-field mdc-text-field--filled mdc-text-field--label-floating">
      <span class="mdc-text-field__ripple"></span>
      <input class="mdc-text-field__input" type="text" aria-labelledby="my-label-id" value="Initial value">
      <span class="mdc-floating-label mdc-floating-label--float-above" id="my-label-id">
        Label in correct place
      </span>
      <span class="mdc-line-ripple"></span>
    </label>
    """
    def __init__(self, name=None, label=None, value=None, **context):
        label = label or (name and name.capitalize()) or ''
        super().__init__(
            Span(cls='mdc-text-field__ripple'),
            Input(
                name=name,
                value=value or '',
                type='text',
                aria_labelledby=f'{name}-label',
                cls='mdc-text-field__input',
                **context,
            ),
            Span(
                Text(label),
                id=f'{name}-label',
                cls='mdc-floating-label',
            ),
            Span(cls='mdc-line-ripple'),
            cls='mdc-text-field mdc-text-field--filled',
            data_mdc_auto_init='MDCTextField',
        )


"""
class Component(Component):
    def __init_(self, *args, **kwargs):
        self.context = kwargs.pop('context', {})
        super().__init_(*args, **kwargs)


class Appbar(Component):
    '''
    Appbar component

    Represents a MUI CSS <appbar>.

    :parameters: see :class:`Component`
    '''
    def __init_(self, content=None, attr=None, events=None,
                 parent='body', id=None, context=None):
        return Div(content,
                   attr={'class': "mui-appbar"},
                   events, parent, id)


class Button(Component):
    '''
    Button component

    Represents a MUI CSS <button>.

    :parameters: see :class:`Component`
    '''
    def __init_(self, content=None, attr=None,
                 events=None, parent='body', id=None, context=None):
        attr = attr or {}
        cls = attr.setdefault('class', '')
        attr['class'] = f"mui-btn {cls}"
        return Button(content, attr, events, parent, id)


class Container(Component):
    '''
    Container component

    Represents a MUI <container>.

    :parameters: see :class:`Component`
    '''
    def __init_(self, content=None, attr=None, events=None,
                 parent='body', id=None, context=None):
        return Div(content,
                   attr={'class': "mui-container"},
                   events, parent, id)
"""


class MdcForm(Component):
    '''
    Form component

    Represents a MUI CSS <form>.

    :parameters: see :class:`Component`
    '''
    def __init_(self, *content, **attrs):
        attrs = {'class': "mui-form"}
        super().__init_(*content, **attrs, tag='form')


class MdcLegend(Component):
    '''
    Legend component

    Represents a MUI CSS <legend>.

    :parameters: see :class:`Component`
    '''
    def __init_(self, *content, **attrs):
        super().__init_(*content, **attrs, tag='legend')



import os

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from django.conf.urls import url
from django.core.asgi import get_asgi_application

from ryzom_django_channels.consumers import Consumer

application = ProtocolTypeRouter({
    # Django's ASGI application to handle traditional HTTP requests
    'http': get_asgi_application(),

    'websocket': AuthMiddlewareStack(
        URLRouter([
            url(r"^ws/ddp/$", Consumer.as_asgi()),
        ])
    ),
})



'''
Ryzom is a reactive full-stack Python framework built on top of the Django
framework.
Ryzom is highly inspired by the open source software MeteorJS; if you like
JavaScript and don't know it, please give it a try!

Ryzom will always stay open source.
'''



from django.core.management.base import BaseCommand, CommandError

from ryzom_django import bundle


class Command(BaseCommand):
    help = 'Output CSS bundle from Ryzom components'

    def handle(self, *args, **options):
        self.stdout.write(bundle.css())



'''
Ryzom components declarations.
There's still a lot of tags missing.
They will be added when they'll be needed
'''
import copy
import importlib
import textwrap
import re
import uuid

from py2js.transpiler import transpile_body


def component_html(path, *args, **kwargs):
    from django.utils.safestring import mark_safe
    try:
        from jinja2.utils import Markup
    except ImportError:
        Markup = None
    import cli2
    ComponentCls = cli2.Node.factory(path).target
    component = ComponentCls(*args, **kwargs)
    html = component.to_html()

    if Markup:
        html = Markup(html)
    return mark_safe(html)


class HTMLPayload(dict):
    def __init__(self, **kwargs):
        super().__init__(**{
            k.replace('_', '-'): v for k, v in kwargs.items()
        })

    def __getattr__(self, name):
        html_name = name.replace('_', '-')
        if html_name in self:
            return self[html_name]
        raise AttributeError(f'{self} object has no attribute {name}')

    def __setattr__(self, name, value):
        self.__setitem__(name, value)

    def __setitem__(self, name, value):
        name = name.replace('_', '-')
        if value is None:
            if name in self:
                del self[name]
        else:
            super().__setitem__(name, value)

    def update(self, other):
        for key, value in other.items():
            self[key] = value


class CStyle(HTMLPayload):
    @classmethod
    def to_dict(cls, value):
        result = {}
        for rule in value.split(';'):
            if not rule.strip():
                continue
            key, value = rule.split(':')
            result[key.strip()] = value.strip()
        return result


class CAttrs(HTMLPayload):
    def __getitem__(self, name):
        if name == 'style' and 'style' not in self:
            # Create CStyle on the fly
            self['style'] = CStyle()

        if name == 'cls':
            # Translate 'cls' into class
            name = 'class'

        return super().__getitem__(name)

    def __getattr__(self, name):
        if name in self or name in ('style', 'cls'):
            return self.__getitem__(name)
        raise AttributeError(f'{self} object has no attribute {name}')

    def __setitem__(self, name, value):
        if name in ('cls', 'class', 'addcls'):
            if isinstance(value, (list, tuple)):
                value = ' '.join(value)
        if name == 'cls':
            # Maintain the "class" attribute
            name = 'class'
        elif name == 'addcls':
            if 'class' not in self:
                self['class'] = value
            elif self['class']:
                self['class'] += ' ' + value
            else:
                self['class'] = value
            return
        elif name == 'rmcls':
            if 'class' in self:
                self['class'] = self['class'].replace(
                    value, ''
                ).replace('  ', ' ').strip()
            return

        if name == 'style' and isinstance(value, str):
            self.style = CStyle.to_dict(value)
        else:
            super().__setitem__(name, value)

    def update(self, other):
        for key, value in other.items():
            if key == 'style':
                if isinstance(value, str):
                    value = CStyle.to_dict(value)
                self['style'].update(value)
            else:
                self[key] = value

    def to_html(self):
        result = []
        for key, value in self.items():
            if key == 'style':
                new_value = []
                for style_key, style_value in value.items():
                    new_value.append(f'{style_key}: {style_value}')
                value = '; '.join(new_value)

            if value is True:
                result.append(key)
            elif value is not False:
                value = str(value).replace('"', '&quot;')
                result.append(f'{key}="{value}"')
        return ' '.join(result)


def noclose(tag):
    return tag.lower() in [
        'area', 'base', 'br', 'col', 'embed', 'hr', 'img',
        'input', 'link', 'meta', 'param', 'source', 'track', 'wbr',
    ]


class ComponentMetaclass(type):
    def __new__(cls, name, bases, class_attrs):
        attrs = CAttrs()
        scripts = list()
        stylesheets = list()

        for base in bases:
            if base_attrs := getattr(base, 'attrs', None):
                attrs.update({k: v for k, v in base_attrs.items() if k != 'style'})
            if extra_scripts := getattr(base, 'scripts', None):
                scripts.extend(extra_scripts)  # scripts are a dict!
            if extra_stylesheets := getattr(base, 'stylesheets', None):
                stylesheets.extend(extra_stylesheets)  # styles are a dict!

        if class_attrs.get('attrs', None):
            attrs.update(class_attrs['attrs'])
        if 'style' in class_attrs:
            attrs.update(dict(style=class_attrs['style']))
        if attrs.get('style', None) or class_attrs.get('sass', None):
            if not attrs.get('class', ''):
                attrs['class'] = name
            else:
                attrs['class'] += f' {name}'
        class_attrs['attrs'] = attrs

        if extra_stylesheets := class_attrs.get('stylesheets', None):
            stylesheets.extend(extra_stylesheets)
        class_attrs['stylesheets'] = stylesheets

        if extra_scripts := class_attrs.get('scripts', None):
            scripts.extend(extra_scripts)
        class_attrs['scripts'] = scripts

        if 'tag' not in class_attrs:
            tag = None
            if 'HTMLElement' not in class_attrs:
                # inherit tag from parent UNLESS I define a WebComponent
                for base in bases:
                    tag = getattr(base, 'tag', None)
                    if tag:
                        break

            if not tag:
                s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
                tag = re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()

            class_attrs['tag'] = tag

        if class_attrs.get('tag', None):
            class_attrs['noclose'] = noclose(class_attrs['tag'])

        cls = super().__new__(cls, name, bases, class_attrs)

        from ryzom.bundle.js import AUTOCOMPILE
        for method in AUTOCOMPILE:
            if getattr(cls, method, None):
                class_attrs['attrs'][method] = f'{name}_{method}(this)'

        return cls


class Component(metaclass=ComponentMetaclass):
    '''Main ryzom component 'abstract' class to be inherited.

    This class defines the common attrsibutes and methods to all
    components,the main one being to_obj() that format an instance
    as a serializable dict that can be sent to the client over websocket
    the to_html() method is missing for now, it will be useful when
    implementing Server-Side Rendering

    Upon creation, if not specified, an instance of the Component class
    is considered as a child of the <html> tag, this is not guaranteed
    to be kept in a near future, because it's totally useless.
    Being a childnode of <body> seem a lot more meaningful.
    If no id is specified, a random (but still unique) one will be
    generated.

    :param str tag: The HTML tag of the component
    :param list<Component> content: The component instances contained \
            by the current instance
    :param dict attrs: HTML attrsibutes (id, class, style, ...)
    :param dict events: The events to add listeners to \
            (click, hover, ...)
    :param str parent: The id of the component that contains the \
            current instance
    :param str id: The id of the current instance (must be unique)
    '''

    tag = None  # make sure this class is tagless
    __publication = None

    def __getattribute__(self, name):
        '''Bind style to attrs.style'''
        if name == 'style':
            return self.attrs.style
        return super().__getattribute__(name)

    def __getattr__(self, name):
        '''Instanciate attrs on the fly.'''
        if name == 'attrs':
            self.attrs = CAttrs()
            return self.attrs
        raise AttributeError(f'{self} object has no attribute {name}')

    def __init__(self, *content, **attrs):
        cls = type(self)
        self.content = list(content) or []

        for key in [*attrs.keys()]:
            value = attrs[key]
            if not hasattr(value, 'to_html'):
                continue
            setattr(self, key, value)
            if hasattr(value, 'attrs'):
                value.attrs.setdefault('slot', key)
            self.content.append(attrs.pop(key))

        self.id = attrs.get('id', uuid.uuid1().hex)
        self.parent = attrs.pop('parent', None)

        if 'tag' in attrs:
            self.__dict__['tag'] = attrs.pop('tag')
            self.__dict__['noclose'] = noclose(self.tag)

        self.__dict__['selfclose'] = attrs.pop(
            'selfclose',
            getattr(self, 'selfclose', False),
        )

        self.scripts = copy.deepcopy(getattr(self, 'scripts', []))
        self.stylesheets = copy.deepcopy(getattr(self, 'stylesheets', []))

        self.events = attrs.pop('events', {})

        class_attrs = copy.deepcopy(getattr(self, 'attrs', {}))
        # remove class defined style attribute because it is bundled
        if 'style' in class_attrs:
            class_attrs.pop('style')

        # create an instance attribute from the class attribute
        self.__dict__['attrs'] = class_attrs

        # update with kwargs
        self.attrs.update(attrs)

        self.position = 0

        self.preparecontent()

    def __eq__(self, other):
        if self.attrs != other.attrs or self.tag != other.tag:
            return False
        for my_content, other_content in zip(self.content, other.content):
            if my_content != other_content:
                return False
        return True

    def preparecontent(self):
        '''Set the parent and position of children

        meant to be called by __init__().
        This method sets the current component as parent of each child
        Moreover it sets the child's position attribute to its index
        in the current component's content list
        '''
        for i, c in enumerate(self.content):
            if c is None:
                del self.content[i]
                i -= 1
                continue

            if not hasattr(c, 'to_html'):
                self.content[i] = c = Text(str(c))

            c.parent = self
            c.position = i

    def addchild(self, component):
        '''Add a child component

        Currently not used at all, but meant to push a new child
        at the end of the content's list
        A call to preparecontent() should follow the call to this method

        :param Component component: The child component to add to the \
                content of the current instance
        '''
        component.position = len(self.content)
        component.parent = self
        self.content.append(component)

    def addchildren(self, components):
        '''Add a list of children

        Currently not used at all, meant to push children at the
        end of the content's list.
        A call to preparecontent() should follow the call to this method

        :param list(Component) components: The component list to insert \
                in the content of the current instance
        '''
        for component in components:
            self.addchild(component)

    def addevents(self, events):
        '''Add events to instance

        Currently not used, meant to add/update a dict of eventListener
        attached to the DOM element associated with this instance

        :param dict events: The dict to update events with
        '''
        self.events.update(events)

    def to_obj(self, context=None):
        '''Get a serializable dict of the instance

        This methods returns a dict representation of the current
        instance. I handles subscriptions that will have this component
        instance id as parent attribute.
        Recursively sets the content as dict too (maybe recursiveness is not
        a good thing to do without any control of how deep can the tree be,
        there's a risk of stack overflow that we must keep in mind)

        :returns: A serializable representation of the instance
        '''
        if self.tag == 'text':
            content = self.content
        else:
            content = []
            for c in self.content:
                if not c:
                    continue
                if isinstance(c, (int, float, str)):
                    content.append(c)
                else:
                    content.append(c.to_obj())

        if isinstance(self.parent, str):
            parent_id = self.parent
        else:
            parent_id = self.parent.id

        return {
            'id': self.id,
            'tag': self.tag,
            'content': content,
            'parent': parent_id,
            'position': self.position,
            'script': self.render_js(),
            'attrs': self.attrs
        }

    @property
    def publication(self):
        return self.__publication

    @publication.setter
    def publication(self, value=None):
        self.__publication = value

    def content_html(self, *content, **context):
        html = ''
        for c in content:
            if hasattr(c, 'to_html'):
                newline = '' if getattr(c, 'tag', None) == 'text' else '\n'
                html += newline + c.to_html(**context)
            else:
                html += str(c)

        return html

    def context(self, *content, **context):
        for c in (content or self.content):
            if hasattr(c, 'context'):
                context = c.context(**context)
        return context

    def to_html(self, *content, attrs=None, **context):
        if self.tag == 'text':
            return f'{self.content}'

        attrs = ' '.join([
            (attrs or self.attrs).to_html(),
            f'ryzom-id="{self.id}"'
        ])
        html = ''

        if getattr(self, 'selfclose', False):
            html = f'<{self.tag} {attrs}/>'
        elif getattr(self, 'noclose', False):
            html = f'<{self.tag} {attrs}>'
        else:
            html = f'<{self.tag} {attrs}>'
            content = content or self.content
            html += self.content_html(*content, **context)
            if render_js_str := self.render_js():
                html += '\n'.join([
                    '\n<script type="text/javascript">',
                    render_js_str.strip(),
                    '</script>',
                ])
            if self.content and getattr(self.content[-1], 'tag', None) != 'text':
                newline = '\n'
            else:
                newline = ''
            html += f'{newline}</{self.tag}>'

        return html

    def render(self, *content, **context):
        if 'view' in context:
            self.view = context['view']
        content = content or self.content
        context = self.context(*content, **context)
        return self.to_html(*content, **context)

    def render_js(self):
        if hasattr(self, 'py2js'):
            return transpile_body(self.py2js, self=self)
        return ''

    def render_js_tree(self, lvl=0):
        js_str = str(self.render_js())

        if js_str:
            js_str = js_str

        if hasattr(self, 'content') and isinstance(self.content, (list, tuple)):
            for c in self.content:
                if isinstance(c, Component):
                    js_str += c.render_js_tree(lvl+1)

        return js_str


class CTree(Component):
    def __init__(self, *components):
        self.components = components
        self.__name__ = components[-1].__name__

    def __call__(self, **kwargs):
        component = self.components[-1](**kwargs)
        for wrapper in reversed(self.components[:-1]):
            component = wrapper(component, **kwargs)
        return component


class CList(Component):
    def to_html(self, **kwargs):
        return self.content_html(*self.content, **kwargs)

    def to_obj(self, context=None):
        content = [
            c.to_obj(context)
            for c in self.content
        ]
        return content


class Text(Component):
    '''
    Text component

    Represents a text node
    '''
    tag = 'text'

    def preparecontent(self):
        self.content = self.content[0]


class Markdown(Text):
    def __init__(self, *content, **kwargs):
        self.kwargs = kwargs
        super().__init__(*content)

    def preparecontent(self):
        import markdown
        from django.utils.safestring import mark_safe
        self.content = mark_safe(
            markdown.markdown(
                textwrap.dedent(
                    '\n'.join(self.content),
                ),
                **self.kwargs,
            )
        )



from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage


class Static:
    '''Return a static url for an app asset.

    :param str src: The app path of the asset.
    '''
    def __init__(self, src):
        self._data = src
        self.url = staticfiles_storage.url(src)

    def __str__(self):
        return self.url



'''
Consumer for Django channels.
Handles websockets messages from client and channels layer
ddp_urlpattern and server_methods are subject to change in a
near future. Both will be handled in a separate file
'''
import importlib
import json
from datetime import timedelta

from asgiref.sync import async_to_sync
from channels.auth import get_user, login
from channels.generic.websocket import JsonWebsocketConsumer
from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth.models import User
from django.utils import timezone

from ryzom_django_channels.methods import Methods
from ryzom_django_channels.models import Client, Publication, Subscription


class Consumer(JsonWebsocketConsumer):
    '''
    Consumer class, inherited from the channels' JsonWebsocketConsumer
    '''
    ddp_urlpatterns = importlib.import_module(
        settings.WS_URLPATTERNS).urlpatterns

    '''
    Import all user defined server methods
    '''
    for module in settings.SERVER_METHODS:
        importlib.import_module(module)

    def connect(self):
        '''
        Websocket connect handler.
        This method tries to get the user connecting and create a new
        ryzom.models.Client in DB, saving the channel name for future
        access from the channel layer.
        sends back a 'Connected' message to the client
        '''
        client = None
        user = async_to_sync(get_user)(self.scope)
        token = self.scope['query_string'].decode()
        if token:
            client = Client.objects.filter(token=token).last()
            if client and client.user:
                async_to_sync(login)(self.scope, client.user)
                self.scope['session'].save()

        self.accept()
        if client and client.channel != self.channel_name:
            client.channel = self.channel_name
            client.user = user if isinstance(user, User) else None
            client.save()
            self.send(json.dumps({'type': 'Connected'}))
        else:
            self.send(json.dumps({'type': 'Reload'}))

    def disconnect(self, close_code):
        '''
        Websocket disconnect handler.
        Removes the ryzom.models.Client entry attached to this
        channel, cascading deletion to Suscriptions
        Zombies that may stay in our DB on server reboots are removed in
        the ryzom.apps Appconfig.ready() function
        '''
        client = Client.objects.filter(channel=self.channel_name)
        if client.count():
            client.delete()

        expiration = timezone.now() - timedelta(minutes=2)
        deadclients = Client.objects.filter(channel='', created__lt=expiration)
        deadclients.delete()

    def receive(self, text_data):
        '''
        Websocket message handler.
        Dispatches message to type specific subhandlers after some
        error checking on the message format
        Known message types are 'subscribe', 'unsubscribe', 'method'
        and 'geturl'.
        In a near future, login and logout could be handled too,
        unless we use another way to do it, by method call or anything else
        A message should have:
        - an 'id' key, which is used to find the right
        callback function the client defined
        - a 'type' key, one of the known message types described above
        - a 'params' key, which is used as a parameter, specific to
        each message type.
        '''
        if not Client.objects.filter(channel=self.channel_name).count():
            self.send(json.dumps({'type': 'Reload'}))
            return

        data = json.loads(text_data)
        msg_type = None
        if not data.get('id', None):
            return
        try:
            msg_type = data['type']
        except KeyError:
            self.send(json.dumps({
                'id': data['id'],
                'type': 'Error',
                'params': {
                    'name': 'Bad message',
                    'message': 'message type not found'
                }
            }))
            return

        if msg_type in [
                'subscribe', 'unsubscribe',
                'method', 'geturl',
                'login', 'logout', 'ping']:
            func = getattr(self, f'recv_{msg_type}', None)
            if func:
                if data.get('params', None) is None:
                    self.send(json.dumps({
                        'id': data.get('id'),
                        'type': 'Error',
                        'params': {
                            'name': 'Bad format',
                            'message': '"params" key not found'
                        }
                    }))
                else:
                    func(data)
        else:
            self.send(json.dumps({
                'id': data['id'],
                'type': 'Error',
                'params': {
                    'name': 'Bad message type',
                    'message': f'{msg_type} not recognized'
                }
            }))

    def recv_ping(self, data):
        self.send(json.dumps({
            'id': data['id'],
            'type': 'pong'
        }))

    def recv_login(self, data):
        credentials = data['params']
        user = authenticate(**credentials)
        if user:
            async_to_sync(login)(self.scope, user)
            self.scope['session'].save()
            client = Client.objects.get(channel=self.channel_name)
            client.user = user
            client.save()
            self.send(json.dumps({
                'id': data['id'],
                'type': 'Success',
                'params': {
                    'token': f'{client.token}'
                }
            }))
        else:
            self.send(json.dumps({
                'id': data['id'],
                'type': 'Error',
                'params': {
                    'name': 'Credentials mismatch',
                    'message': 'Wrong username/password combination'
                }
            }))

    def recv_logout(self, data):
        pass

    def recv_geturl(self, data):
        '''
        geturl message handler.
        Creates a new ryzom.views.View based on ddp_urlpattern configuration
        and attach it to this consumer instance.
        Renders the view then send it to the client
        If a view as already been created, destroy it and creates the new one
        view's callback (oncreate, ondestroy) are called here
        '''
        to_url = data['params'].get('url', '/')
        to_query = data['params'].get('query', '')
        for url in Consumer.ddp_urlpatterns:
            if url.pattern.match(to_url):
                cview = getattr(self, 'view', None)
                if not cview or not isinstance(cview, url.callback):
                    if cview:
                        cview.ondestroy()
                    client = Client.objects.filter(channel=self.channel_name).last()
                    req = Request(client, url.callback)
                    cview = self.view = url.callback(req)
                    cview.oncreate(to_url)
                if (cview.onurl(to_url)):
                    self.send(json.dumps({
                        'id': data['id'],
                        'type': 'Success',
                        'params': []
                    }))
                else:
                    data = {
                        'id': data['id'],
                        'type': 'Error',
                        'params': ''
                    }
                    self.send(json.dumps(data))
                break

    def recv_method(self, data):
        '''
        method message handler.
        Lookup methods then call them with the 'params' key as parameter.
        Methods writers should handle that params.
        Methods should return a value that evaluates to True on Success.
        Methods return value should be serializable, it will be sent
        to the client as parameter for the callback
        '''
        to_send = {'id': data['id']}
        params = data['params']
        method = Methods.get(params['name'])
        if method is None:
            to_send.update({
                'type': 'Error',
                'params': {
                    'name': 'Not found',
                    'message': f'Method {params["name"]} not found'
                }
            })
        else:
            user = async_to_sync(get_user)(self.scope)
            ret = method(user, params['params'])
            if ret:
                to_send.update({
                    'type': 'Success',
                    'params': ret
                })
            else:
                to_send.update({
                    'type': 'Error',
                    'params': ret
                })
        self.send(json.dumps(to_send))

    def insert_component(self, data, change=False):
        '''
        This method is meant to be called by the DDP dispacher.
        It send a DDP insert/change message to the client with
        a serialized component as params
        '''
        self.send(json.dumps({
            'type': 'DDP',
            'params': {
                'type': 'insert' if not change else 'change',
                'params': data['instance']
            }
        }))

    def remove_component(self, data):
        '''
        This method is meant to be called by the DDP dispacher.
        It send a DDP remove message to the client with the parent
        and id of the component to remove as params
        '''
        self.send(json.dumps({
            'type': 'DDP',
            'params': {
                'type': 'remove',
                'params': {
                    'id': data['id'],
                    'parent': data['parent']
                }
            }
        }))

    def handle_ddp(self, data):
        '''
        DDP dispacher.
        handler for 'handle.ddp' messages sent over the channel layer.
        dispaches the message to the above two methods
        '''
        if data['params']['type'] == 'inserted':
            self.insert_component(data['params'])
        elif data['params']['type'] == 'changed':
            self.insert_component(data['params'], True)
        elif data['params']['type'] == 'removed':
            self.remove_component(data['params'])

    def recv_subscribe(self, data):
        '''
        subscribe message handler.
        Creates a new subscription for the current Client.
        'subscribe' params should contain:
        - an 'id' key, which refer to the component that asks for
        a subscription
        - a 'name' key, corresponding to the name of the publication
        this subscription is about
        '''
        params = data['params']
        to_send = {'id': data['id']}
        client = Client.objects.get(channel=self.channel_name)
        print(f'GOT SUBSCRIBE FOR CLIENT {client}')
        for key in ['name', 'sub_id']:
            if key not in params:
                to_send.update({
                    'type': 'Error',
                    'params': {
                        'name': 'Bad format',
                        'message': f'Subscription {key} not found'
                    }
                })
                self.send(json.dumps(to_send))
                return
        if not client:
            to_send.update({
                'type': 'Error',
                'params': {
                    'name': 'Client not found',
                    'message': 'No client was found for this channel name'
                }
            })
        else:
            pub = Publication.objects.get(name=params['name'])
            sub = Subscription.objects.filter(
                publication=pub,
                parent=params['parent_id'],
                client=client).first()
            if not sub:
                sub = Subscription(
                        publication=pub,
                        parent=params['parent_id'],
                        client=client)
                sub.save()
            else:
                print('SUBSCRIPTION FOUND')
                sub.exec_query(params['opts'])
            print(sub.client, sub)

            to_send.update({
                'type': 'Success',
                'params': {
                    'name': params['name'],
                    'sub_id': f"{sub.id}",
                    'length': len(sub.queryset)
                }
            })
        self.send(json.dumps(to_send))

    def recv_unsubscribe(self, data):
        '''
        unsubscribe message handler.
        not implemented yes but meant to delete the subscription
        attached to the current Client/Publication name from DB
        '''
        params = data['params']
        self.send(json.dumps({
            'id': data['id'],
            'type': 'unsubscribed',
            'message': 'Got unsub',
            'params': {
                'name': params['name']
            }
        }))



import importlib
import sys

from django import http
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.views import generic
from django.urls import include, path, reverse_lazy

from ryzom import bundle
from ryzom import html


class CSSBundle(html.Stylesheet):
    def to_html(self, *content, **context):
        if settings.DEBUG:
            self.attrs.href = reverse_lazy('bundle_css')
        else:
            self.attrs.href = staticfiles_storage.url('bundle.css')
        return super().to_html(*content, **context)


class JSBundle(html.Script):
    def to_html(self, *content, **context):
        if settings.DEBUG:
            self.attrs.src = reverse_lazy('bundle_js')
        else:
            self.attrs.src = staticfiles_storage.url('bundle.js')
        return super().to_html(*content, **context)


def get_component_modules():
    names = []
    for app in settings.INSTALLED_APPS:
        for subname in ('views', 'urls', 'components', 'html'):
            name = '.'.join([app, subname])
            try:
                importlib.import_module(name)
            except ImportError:
                continue
            names.append(name)
    return names


def js():
    return bundle.js(*get_component_modules())


def css():
    return bundle.css(*get_component_modules())


class CSSBundleView(generic.View):
    def get(self, *args, **kwargs):
        response = http.HttpResponse(css())
        response['Content-Type'] = 'text/css'
        return response

    @classmethod
    def as_url(cls):
        return path('bundle.css', cls.as_view(), name='bundle_css')


class JSBundleView(generic.View):
    def get(self, *args, **kwargs):
        response = http.HttpResponse(js())
        response['Content-Type'] = 'text/javascript'
        return response

    @classmethod
    def as_url(cls):
        return path('bundle.js', cls.as_view(), name='bundle_js')


urlpatterns = [
    JSBundleView.as_url(),
    CSSBundleView.as_url(),
]





class UnpolyMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        response['X-Up-Location'] = request.path_info + '?' + request.GET.urlencode()
        return response



from django.conf import settings
from django.urls import include, path

urlpatterns = [
    path('', include('ryzom_django_example.views')),
]

if settings.DEBUG:
    urlpatterns += [path('bundles/', include('ryzom_django.bundle'))]

if settings.CHANNELS_ENABLE:
    urlpatterns.append(
        path('reactive/', include('ryzom_django_channels_example.views')),
    )



import os
import socket
from pathlib import Path

REDIS_SERVER = None
CHANNELS_ENABLE = False

REDIS_SERVERS = [
    ('redis', 6379),
    ('127.0.0.1', 6379)
]

if 'CHANNELS_ENABLE' in os.environ:
    CHANNELS_ENABLE = bool(os.environ['CHANNELS_ENABLE'])

for server in REDIS_SERVERS:
    a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        result_of_check = a_socket.connect_ex(server)
    except socket.gaierror:
        continue

    if result_of_check == 0:
        REDIS_SERVER = server
        CHANNELS_ENABLE = True
        break

BASE_DIR = Path(__file__).resolve().parent.parent
SECRET_KEY = '4am4pn_87&v0qaq%_-2me06et#@prq(yp6npk8g495!@7s1hoi'
DEBUG = True
ALLOWED_HOSTS = []

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'ryzom_django_example',

    # Enable components templates auto discover
    'ryzom_django',

    # Add py2js static file
    'py2js',

    # Enable form rendering with MDC components
    'ryzom_django_mdc',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]


# Enable Ryzom template backend
TEMPLATES = [
    {
        'BACKEND': 'ryzom_django.template_backend.Ryzom',
        'NAME': 'ryzom',
    },
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

if CHANNELS_ENABLE:
    # Enable Reactive components models
    INSTALLED_APPS += [
        'ryzom_django_channels',
        'ryzom_django_channels_example',
        'channels',
        'channels_redis',
    ]

ROOT_URLCONF = 'ryzom_django_example.urls'
WS_URLPATTERNS = ROOT_URLCONF
SERVER_METHODS = []

ASGI_APPLICATION = 'ryzom_django_example.asgi.application'

if CHANNELS_ENABLE:
    CHANNEL_LAYERS = {
        "default": {
            "BACKEND": "channels_redis.core.RedisChannelLayer",
            "CONFIG": {
                "hosts": [REDIS_SERVER],
            },
        },
    }


DATABASES = {
    'default': {
        'ENGINE': os.getenv('DB_ENGINE', 'django.db.backends.postgresql'),
        'NAME': os.getenv('DB_NAME', 'ryzom_django_example'),
        'HOST': os.getenv('DB_HOST'),
        'USER': os.getenv('DB_USER'),
        'PASSWORD': os.getenv('DB_PASSWORD'),
    }
}


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = BASE_DIR / 'static'
